#!/usr/bin/env python
"""Collection of functions to access, handle and manipulate different entities of
COBRApy and libsbml models.
"""

__author__ = "Famke Baeuerle and Gwendolyn O. Döbel and Carolin Brune"

################################################################################
# requirements
################################################################################

import ast
import cobra 
import json
import pandas as pd
import re
import requests
import urllib
import warnings

from Bio import Entrez
from Bio.KEGG import REST, Compound
from libsbml import Model as libModel
from libsbml import GeneProduct, Species, Reaction, FbcOr, FbcAnd, GeneProductRef

from random import choice
from string import ascii_uppercase, digits
from typing import Union, Literal
from tqdm import tqdm

tqdm.pandas()
pd.options.mode.chained_assignment = None # suppresses the pandas SettingWithCopyWarning; comment out before developing!!

from .cvterms import add_cv_term_genes, add_cv_term_metabolites, add_cv_term_reactions, _add_annotations_from_dict_cobra
from .db_access import search_ncbi_for_gpr, kegg_reaction_parser, _add_annotations_from_bigg_reac_row, get_BiGG_metabs_annot_via_dbid, add_annotations_from_BiGG_metabs
from .io import load_a_table_from_database
from .util import COMP_MAPPING, VALID_COMPARTMENTS
from ..developement.decorators import *

################################################################################
# variables
################################################################################


################################################################################
# functions
################################################################################

# ++++++++++++++++++++++++++++++++++++++++
# COBRApy - models
# ++++++++++++++++++++++++++++++++++++++++

# handling compartments 
# ---------------------

def are_compartment_names_valid(model:cobra.Model) -> bool:
    """Check if compartment names of model are considered valid based on 
    :py:const:`~refinegems.utility.util.VALID_COMPARTMENTS`.

    Args:
        - model (cobra.Model): 
            The model, loaded with COBRApy.

    Returns:
        bool: 
            True, if valid, else false.
    """

    for c in model.compartments.keys():
        if c not in VALID_COMPARTMENTS.keys():
            return False

    return True


# @TODO extend + change else case?
def resolve_compartment_names(model:cobra.Model):
    """Resolves compartment naming problems.

    Args:
        - model (cobra.Model): 
            A COBRApy model object.

    Raises:
        - KeyError: Unknown compartment raises an error to add it to the mapping. Important for developers.
    """

    # check if compartment names are valid
    if not are_compartment_names_valid(model):

        # check if mapping is possible
        if set(model.compartments.keys()).issubset(set(COMP_MAPPING.keys())):
            # for each metabolite rename the compartment
            for metabolite in model.metabolites:
                metabolite.compartment = COMP_MAPPING[metabolite.compartment]
            # add whole descriptions of the compartments to the model
            # note:
            #    only compartments IN the model will be added
            model.compartments = VALID_COMPARTMENTS

        else:
            raise KeyError(F'Unknown compartment {[_ for _ in model.compartments if _ not in COMP_MAPPING.keys()]} detected. Cannot resolve problem.')


# handling cobra entities (features)
# ----------------------------------
        
def reaction_equation_to_dict(eq: str, model: cobra.Model) -> dict:
    """Parses a reaction equation string to dictionary 

    Args:
        - eq (str): 
            Equation of a reaction
        - model (cobra.Model): 
            Model loaded with COBRApy
        
    Returns:
        dict: 
            Metabolite Ids as keys and their coefficients as values (negative = reactants, positive = products)
    """
    # from Alina Renz
    eq = eq.split(' ')
    eq_matrix={}
    are_products = False
    coeff = 1
    for i,part in enumerate(eq):
        if part == '-->':
            are_products = True
            continue          
        if part == '+':
            continue
        if part == '2':
            coeff = 2
            continue
        if are_products:
            eq_matrix[model.metabolites.get_by_id(part)] = 1*coeff
            coeff = 1
        else:
            eq_matrix[model.metabolites.get_by_id(part)] = -1*coeff
            coeff = 1
    return eq_matrix


def get_reaction_annotation_dict(model:cobra.Model, db:Literal['KEGG','BiGG']) -> dict:
    """Create a dictionary of a model's reaction IDs and a choosen database ID as
    saved in the annotations of the model.

    The database ID can be choosen based on the strings for the namespace options
    in other functions.

    Args:
        - model (cobra.Model): 
            A model loaded with COBRApy.
        - db (Literal['KEGG','BiGG']): 
            The string denoting the database to map to.

    Raises:
        - ValueError: Unknown database string for paramezer db

    Returns:
        dict: 
            The mapping of the reaction IDs to the database IDs found in the annotations
    """

    react_dict = {}

    match db:
        case 'KEGG':
            db_string = 'kegg.reaction'
        case 'BiGG':
            db_string = 'bigg.reaction'
        case _:
            mes = f'Unknown database string for parameter db: {db}'
            raise ValueError(mes)

    for r in model.reactions:
        if db_string in r.annotation.keys():
            react_dict[r.id] = r.annotation[db_string]
        else:
            react_dict[r.id] = '-'

    return react_dict


def create_random_id(model:cobra.Model, entity_type:Literal['reac','meta']='reac', prefix:str='') -> str:
    """Generate a unique, random ID for a model entity for a model.

    Args:
        - model (cobra.Model): 
            A model loaded with COBRApy.
        - entity_type (Literal['reac','meta'], optional): 
            Type of model entity.  
            Can be 'reac' for Reaction or 'meta' for Metabolite.
            Defaults to 'reac'.
        - prefix (str, optional): 
            Prefix to set for the randomised part.
            Useful to identify the random IDs later on. 
            Defaults to ''.

    Raises:
        - ValueError: Unknown entity_type

    Returns:
        str: 
            The generate new and unique ID.
    """

    match entity_type:
        case 'reac':
            all_ids = [_.id for _ in model.reactions]
        case 'meta':
            all_ids = [_.id for _ in model.metabolites]
        case _:
            mes = f'Unkown entity_type: {entity_type}'
            raise ValueError(mes)

    prefix = f'{prefix}{entity_type}'
    var = ''.join(choice(ascii_uppercase + digits) for i in range(6))
    label = prefix + var
    j = 6
    
    while True:
        
        for i in range(36**6): # make sure it does not run endlessly
            if label in all_ids:
                label = prefix + ''.join(choice(ascii_uppercase + digits) for x in range(j))
            else:
                return label
            
        j = j + 1


# @NOTE: Add to reaction handling for `build_reac_bigg`?
def extract_metabolites_from_reactions(missing_reactions: pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame]:
   """Extracts a set of all reactants & products from the missing reactions
   
   Args:
      - missing_reactions (pd.DataFrame): 
         Table containing all missing reactions found through the 
         missing genes
                                          
   Returns:
      tuple: 
         Two tables (1) & (2) 

         (1) pd.DataFrame: Table with the column Compound containing all compounds required for the missing BioCyc reactions
         (2) pd.DataFrame: Table with the column bigg_id containing all compounds required for the missing BiGG reactions
   """
   # Get all BioCyc metabolites necessary for the BioCyc reactions
   biocyc_reactants = [r for row in missing_reactions['Reactants'] for r in row]
   biocyc_products = [p for row in missing_reactions['Products'] for p in row]
   biocyc_metabolites = list(set([*biocyc_reactants, *biocyc_products]))
   
   # Get all BiGG metabolites necessary for the BiGG reactions
   bigg_reactions = [ast.literal_eval(str(reac_dict)) for reac_dict in missing_reactions['bigg_reaction']]
   bigg_reactants = [r for reac_dict in bigg_reactions for r in reac_dict.get('reactants')]
   bigg_products = [p for reac_dict in bigg_reactions for p in reac_dict.get('products')]
   bigg_metabolites = list(set([*bigg_reactants, *bigg_products]))
   
   return (pd.DataFrame(biocyc_metabolites, columns=['Compound']), pd.DataFrame(bigg_metabolites, columns=['bigg_id']))


# @TODO: 
#     more namespace options
#     Maybe bioregistry for mapping of id to namespace?
def match_id_to_namespace(model_entity:Union[cobra.Reaction, cobra.Metabolite], namespace:Literal['BiGG']) -> None:
    """Based on a given namespace, change the ID of a given model entity to it the set namespace.

    Currently working namespaces:

        - BiGG 

    Args:
        - model_entity (cobra.Reaction, cobra.Metabolite]): 
            The model entity. 
            Can be either a cobra.Reaction or cobra.Metabolite object.
        - namespace (Literal['BiGG']): 
            The chosen namespace.

    Raises:
        - ValueError: Unknown input for namespace
        - TypeError: Unknown type for model_entity
    """

    match model_entity:

        # Reaction
        # --------
        case cobra.Reaction():
            match namespace:

                case 'BiGG':
                    if 'bigg.reaction' in model_entity.annotation.keys():
                        # @TODO : currently takes first entry is annotation is list
                        model_entity.id = model_entity.annotation['bigg.reaction'] if isinstance(model_entity.annotation['bigg.reaction'],str) else model_entity.annotation['bigg.reaction'][0]

                case _:
                    mes = f'Unknown input for namespace: {namespace}'
                    raise ValueError(mes)
                
        # Metabolite
        # ----------
        case cobra.Metabolite():
            match namespace:

                case 'BiGG':
                    if 'bigg.metabolite' in model_entity.annotation.keys():
                        model_entity.id = model_entity.annotation['bigg.metabolite'] + '_' + model_entity.compartment if isinstance(model_entity.annotation['bigg.metabolite'],str) else model_entity.annotation['bigg.metabolite'][0]

                case _:
                    mes = f'Unknown input for namespace: {namespace}'
                    raise ValueError(mes)
        # Error
        # -----
        case _:
            mes = f'Unknown type for model_entity: {type(model_entity)}'
            raise TypeError(mes)
        

# originally from SPECIMEN HQTB --- changed / extended
# @TODO - are all needed checks covered? any more needed?
def isreaction_complete(reac:cobra.Reaction, 
                        name_check:bool=False,
                        formula_check:Literal['none','existence','wildcard','strict']='existence',
                        exclude_dna:bool=True, 
                        exclude_rna:bool=True) -> bool:
    """Check, if a reaction object can be considered a complete reaction.
    The parameters can set the strictness of the checking.
    Useful for checking, if a reaction can be added to a model 
    or if it might break it. For example, a missing model ID in either reaction or
    metabolites returns false.
    
    

    Args:
        - reac (cobra.Reaction): 
            The reaction object (COBRApy) to test.
        - name_check (bool, optional): 
            Option to force reaction and metabolites to have the name attribute set. 
            Defaults to False.
        - formula_check (Literal['none','existence','wildcard','strict'], optional): 
            Option to check the formula. 'none' disables the check, 
            'existence' tests, if a formula is set, 'wildcard' additionally checks for wild cards
            (returns false if one found) and 'strict' also checks for the rest symbol 'R'.
            Defaults to 'existence'.
        - exclude_dna (bool, optional): 
            Option to set DNA reactions to invalid. Defaults to True.
        - exclude_rna (bool, optional): 
            Option to set RNA reaction to invalid. Defaults to True.

    Returns:
        bool: 
            True, if the checks are passed successfully, else False.
    """
    
    # check reaction features
    # -----------------------
    # check name
    if name_check:
        if reac.id == '' or pd.isnull(reac.name):
            return False
    # check for RNA/DNA
    if exclude_dna and 'DNA' in reac.name:
        return False
    if exclude_rna and 'RNA' in reac.name:
        return False

    # check metabolites features
    # --------------------------
    for m in reac.metabolites:
        # check id
        if m.id == '' or pd.isnull(m.id):
            return False
        # check name
        if name_check and (m.name == '' or pd.isnull(m.name)):
            return False
        # check formula
        match formula_check:
            # case 1: no formula checking
            case 'none':
                pass
            # case 2: check, if formula is set
            case 'existence':
                if m.formula == '' or pd.isnull(m.formula):
                    return False
            # case 3: check, if formula is set and only contains alphanumerical characters
            case 'wildcard':
                if m.formula == '' or pd.isnull(m.formula) or not m.formula.isalnum():
                    return False
            # case 4: check, if formula is set, constains only alphanumerical characters and no rest R
            case 'strict':
                if m.formula == '' or pd.isnull(m.formula) or not m.formula.isalnum() or bool(re.search(r'R(?![a-z])', 'CH2HOROH')):
                    return False
            # default case: Warning + no formula check
            case _:
                mes = f'Unknown options for formula_check: {formula_check}\nChecking the metabolite formula will be skipped.'
                warnings.warn(mes,UserWarning)
        

    return True


# adding metabolites to cobra models
# ----------------------------------

@template
def build_metabolite_xxx(id:str, model:cobra.Model, 
                         namespace:str,
                         compartment:str,
                         idprefix:str) -> cobra.Metabolite: 
    """Template function for building a cobra.Metabolite.
    
    .. note::
    
        This is a template function for developers. It cannot be executed.

    Args:
        - id (str): 
            _description_
        - model (cobra.Model): 
            _description_
        - namespace (str): 
            _description_
        - compartment (str): 
            _description_
        - idprefix (str): 
            _description_

    Returns:
        cobra.Metabolite: 
            _description_
    """
    # change xxx to database or way the metabolite will be reconstructed with
    # check if id in model
    # get information via id
    # collection formation in a new metabolite object
    # add more annotations from other databases
    # adjust namespace
    # check model again for new namespace
    pass


# originally from SPECIMEN
# @TODO some issues left
# current version works on a couple of examples 
def build_metabolite_mnx(id: str, model:cobra.Model, 
                         namespace:str='BiGG',
                         compartment:str='c',
                         idprefix:str='refineGEMs') -> cobra.Metabolite | None:
    """Build a cobra.Metabolite object from a MetaNetX ID. 
    This function will NOT directly add the metabolite to the model, 
    if the contruction is successful.

    Args:
        - id (str): 
            A MetaNetX ID of a metabolite.
        - model (cobra.Model): 
            The model, the metabolite will be build for.
        - namespace (str, optional): 
            Name to use for the model ID. If namespace cannot be matched,
            will use a random ID. 
            Defaults to 'BiGG'.
        - compartment (str, optional): 
            Compartment of the metabolite. Defaults to 'c'.
        - idprefix (str, optional): 
            Prefix for the random ID. Defaults to 'refineGEMs'.

    Returns:
        1. Case construction successful or match found in model:
                cobra.Metabolite: 
                    The metabolite object.
        
        2. Case construction failed:      
                None:
                    Nothing to return.
    """

    # fast check if compound already in model
    # ------------------------------------------
    # step 1: check if MetaNetX ID in model
    matches = [x.id for x in model.metabolites if 'metanetx.chemical' in x.annotation and x.annotation['metanetx.chemical']==id and x.compartment == compartment]

    # step 2: if yes, retrieve metabolite from model
        # case 1: multiple matches found
    if len(matches) > 0:
        if len(matches) > 1:
            # ................
            # @TODO what to do
            # currently, just the first one is taken
            # ................
            match = model.metabolites.get_by_id(matches[0])
        #  case 2: only one match found
        else:
            match = model.metabolites.get_by_id(matches[0])

        # step 3: add metabolite
        return match

    # if not, create new metabolite
    # -----------------------------
    metabolite_prop = load_a_table_from_database(f'SELECT * FROM mnx_chem_prop WHERE id = \'{id}\'')
    metabolite_anno = load_a_table_from_database(f'SELECT * FROM mnx_chem_xref WHERE id = \'{id}\'')
    if len(metabolite_prop) == 0: # cannot construct metabolite
        return None
    else:
        
        # step 1: create a random metabolite ID
        new_metabolite = cobra.Metabolite(create_random_id(model, 'meta', idprefix)) 

        # step 2: add features
        # --------------------
        new_metabolite.formula = metabolite_prop['formula'].iloc[0]
        new_metabolite.name = metabolite_prop['name'].iloc[0]
        new_metabolite.charge = metabolite_prop['charge'].iloc[0]
        new_metabolite.compartment = compartment

        # step 3: add notes
        # -----------------
        new_metabolite.notes['created with'] = 'refineGEMs GapFiller, metanetx.chemical'

        # step 4: add annotations
        # -----------------------
        # add SBOTerm
        new_metabolite.annotation['sbo'] = 'SBO:0000247'
        
        # add information directly available from the mnx_chem_prop table 
        new_metabolite.annotation['metanetx.chemical'] = [metabolite_prop['id'].iloc[0]]
        if not pd.isnull(metabolite_prop['InChIKey'].iloc[0]):
            new_metabolite.annotation['inchikey'] = metabolite_prop['InChIKey'].iloc[0].split('=')[1]
        
        # get more annotation from the mnx_chem_xref table
        for db in ['kegg.compound','metacyc.compound','seed.compound','bigg.metabolite','chebi']:
            db_matches = metabolite_anno[metabolite_anno['source'].str.contains(db)]
            if len(db_matches) > 0:
                new_metabolite.annotation[db] = [m.split(':',1)[1] for m in db_matches['source'].tolist()]

        # Cleanup BiGG annotations (MetaNetX only saves universal)
        # @TODO : there is no guarantee, that the id with the specific compartment actually exists -> still do it? // kepp the universal id?
        # @TODO : save ID in the correct way
        if 'bigg.metabolite' in new_metabolite.annotation.keys():
            new_metabolite.annotation['bigg.metabolite'] = [_+'_'+compartment for _ in new_metabolite.annotation['bigg.metabolite']]
        else:
            # if no BiGG was found in MetaNetX, try reverse search in BiGG
            get_BiGG_metabs_annot_via_dbid(new_metabolite, id, 'MetaNetX (MNX) Chemical', compartment)
                
        # add additional information from BiGG (if ID found)    
        add_annotations_from_BiGG_metabs(new_metabolite)

        # step 5: change ID according to namespace
        # ----------------------------------------
        match_id_to_namespace(new_metabolite,namespace)
       
        # step 6: re-check existence of ID in model
        # -----------------------------------------
        # @TODO : check complete annotations? 
        #        - or let those be covered by the duplicate check later on?
        if new_metabolite.id in [_.id for _ in model.metabolites]:
            return model.metabolites.get_by_id(new_metabolite.id)
           
    return new_metabolite


# originally from SPECIMEN
# @TODO some issues left
# current version works on a couple of examples 
def build_metabolite_kegg(kegg_id:str, model:cobra.Model, 
                          namespace:Literal['BiGG']='BiGG', 
                          compartment:str='c',
                          idprefix='refineGEMs') -> cobra.Metabolite | None:
    """Build a cobra.Metabolite object from a KEGG ID. 
    This function will NOT directly add the metabolite to the model, 
    if the contruction is successful.

    Args:
        - kegg_id (str): 
            A KEGG ID of a metabolite.
        - model (cobra.Model): 
            The model, the metabolite will be build for.
        - namespace (str, optional): 
            Name to use for the model ID. If namespace cannot be matched,
            will use a random ID. 
            Defaults to 'BiGG'.
        - compartment (str, optional): 
            Compartment of the metabolite. Defaults to 'c'.
        - idprefix (str, optional): 
            Prefix for the random ID. Defaults to 'refineGEMs'.

    Returns:
        1. Case construction successful or match found in model:     
                cobra.Metabolite: 
                    The build metabolite object.
        
        2. Case construction failed:
                None:
                    Nothing to return.
    """
    
    # ---------------------------------------
    # fast check if compound already in model
    # ---------------------------------------
    # step 1: check via KEGG ID
    matches = [x.id for x in model.metabolites if ('kegg.compound' in x.annotation and x.annotation['kegg.compound'] == kegg_id)]
    if len(matches) > 0:
        # step 2: model id --> metabolite object
        #  case 1: multiple matches found
        if len(matches) > 1:
            # .......
            # @TODO
            # .......
            match = model.metabolites.get_by_id(matches[0])
        #  case 2: only one match found
        else:
            match = model.metabolites.get_by_id(matches[0])

        # step 3: add metabolite
        return match

    # -----------------------------
    # if not, create new metabolite
    # -----------------------------
    
    # step 1: retrieve KEGG entry for compound
    # ----------------------------------------
    try:
        kegg_handle = REST.kegg_get(kegg_id)
        kegg_record = [r for r in Compound.parse(kegg_handle)][0]
    except urllib.error.HTTPError:
        warnings.warn(F'HTTPError: {kegg_id}')
        return None
    except ConnectionResetError:
        warnings.warn(F'ConnectionResetError: {kegg_id}')
        return None
    except urllib.error.URLError:
        warnings.warn(F'URLError: {kegg_id}')
        return None

    # step 2: create a random metabolite ID
    # -------------------------------------
    new_metabolite = cobra.Metabolite(create_random_id(model, 'meta',idprefix)) 

    # step 3: add features
    # --------------------
    # set name from KEGG and additionally use it as ID if there is none yet
    if isinstance(kegg_record.name, list) and len(kegg_record.name) > 0:
        # @DISCUSSION : better way to choose a name than to just take the first entry???
        new_metabolite.name = kegg_record.name[0]
    elif isinstance(kegg_record.name, str) and len(kegg_record.name) > 0:
        new_metabolite.name = kegg_record.name
    else:
        new_metabolite.name = '' # @DISCUSSION any ideas for good default values?
    # set compartment
    new_metabolite.compartment = compartment
    # set formula
    new_metabolite.formula = kegg_record.formula

    # step 4: add notes
    # -----------------
    new_metabolite.notes['created with'] = 'refineGEMs GapFiller, KEGG.compound'

    # step 5: add annotations
    # -----------------------
    # add annotation from the KEGG entry
    new_metabolite.annotation['kegg.compound'] = kegg_id
    db_idtf = {'CAS':'cas','PubChem':'pubchem.compound','ChEBI':'chebi'}
    for db,ids in kegg_record.dblinks:
        if db in db_idtf:
            new_metabolite.annotation[db_idtf[db]] = ids
            
    # add SBOTerm
    new_metabolite.annotation['sbo'] = 'SBO:0000247'

    # search for infos in MetaNetX
    mnx_info = load_a_table_from_database(
        f'SELECT * FROM mnx_chem_xref WHERE source = \'kegg.compound:{kegg_id}\'',
        query=True
    )
    # if matches have been found
    if len(mnx_info) > 0:
        mnx_ids = list(set(mnx_info['id']))
        # mapping is unambiguously
        if len(mnx_ids) == 1:
            mnx_info = load_a_table_from_database(
            f'SELECT * FROM mnx_chem_prop WHERE id = \'{mnx_ids[0]}\'',
            query=True
            )
            # add charge 
            new_metabolite.charge = mnx_info['charge'].iloc[0]
            # add more annotations
            new_metabolite.annotation['metanetx.chemical'] = [mnx_info['id'].iloc[0]]
            if not pd.isnull(mnx_info['InChIKey'].iloc[0]):
                new_metabolite.annotation['inchikey'] = mnx_info['InChIKey'].iloc[0].split('=')[1]
            
            # get more annotation from the mnx_chem_xref table 
            metabolite_anno = load_a_table_from_database(f'SELECT * FROM mnx_chem_xref WHERE id = \'{mnx_info["id"]}\'')
            for db in ['kegg.compound','metacyc.compound','seed.compound','bigg.metabolite','chebi']:
                db_matches = metabolite_anno[metabolite_anno['source'].str.contains(db)]
                if len(db_matches) > 0:
                    mnx_tmp = [m.split(':',1)[1] for m in db_matches['source'].tolist()]
                    if db in new_metabolite.annotation.keys():
                        new_metabolite.annotation[db] = list(set(mnx_tmp + new_metabolite.annotation[db]))
                    else:
                        new_metabolite.annotation[db] = mnx_tmp

        else:
            pass
            # @TODO : how to handle multiple matches, e.g. getting charge will be complicated
        
    # Cleanup BiGG annotations (MetaNetX only saves universal)
    # @TODO : there is no guarantee, that the id with the specific compartment actually exists -> still do it? // kepp the universal id?
    if 'bigg.metabolite' in new_metabolite.annotation.keys():
        new_metabolite.annotation['bigg.metabolite'] = [_+'_'+compartment for _ in new_metabolite.annotation['bigg.metabolite']]
    
    # if no BiGG ID, try reverse search
    get_BiGG_metabs_annot_via_dbid(new_metabolite, id, 'KEGG Compound', compartment)
    
    # search for annotations in BiGG
    add_annotations_from_BiGG_metabs(new_metabolite)

    # step 6: change ID according to namespace
    # ----------------------------------------
    match_id_to_namespace(new_metabolite,namespace)
    
    # step 7: re-check existence of ID in model
    # -----------------------------------------
    # @TODO : check complete annotations? 
    #        - or let those be covered by the duplicate check later on?
    if new_metabolite.id in [_.id for _ in model.metabolites]:
        return model.metabolites.get_by_id(new_metabolite.id)

    return new_metabolite


# @TEST some more, somewhat works, but who knows...
# @TODO some comments inside
# @NOTE expects the non-universal BiGG ID (meaning the one with the compartment abbreviation
#       at the end) -> change behaviour or keep it?
def build_metabolite_bigg(id:str, model:cobra.Model, 
                         namespace:Literal['BiGG']='BiGG',
                         idprefix:str='refineGEMs') -> cobra.Metabolite | None: 
    """Build a cobra.Metabolite object from a BiGG ID. 
    This function will NOT directly add the metabolite to the model, 
    if the contruction is successful.

    Args:
        - id (str): 
            A BiGG ID of a metabolite.
        - model (cobra.Model): 
            The model, the metabolite will be build for.
        - namespace (str, optional): 
            Name to use for the model ID. If namespace cannot be matched,
            will use a random ID. 
            Defaults to 'BiGG'.
        - compartment (str, optional): 
            Compartment of the metabolite. Defaults to 'c'.
        - idprefix (str, optional): 
            Prefix for the random ID. Defaults to 'refineGEMs'.

    Returns:
        1. Case construction successful or match found in model:
                cobra.Metabolite: 
                    The build metabolite object.
        
        2. Case construction failed:   
                None:
                    Nothing to return.
    """

    
    compartment = id.rsplit('_',1)[1]
    # ------------------------------------------
    # fast check if compound already in model
    # ------------------------------------------
    # step 1: check if MetaNetX ID in model
    matches = [x.id for x in model.metabolites if 'bigg.metabolite' in x.annotation and (x.annotation['bigg.metabolite']==id or x.annotation['bigg.metabolite']==id.rsplit('_',1)[0]) and x.compartment == compartment]
    # step 2: if yes, retrieve metabolite from model
        # case 1: multiple matches found
    if len(matches) > 0:
        if len(matches) > 1:
            # ................
            # @TODO what to do
            # currently, just the first one is taken
            # ................
            match = model.metabolites.get_by_id(matches[0])
        #  case 2: only one match found
        else:
            match = model.metabolites.get_by_id(matches[0])

        # step 3: add metabolite
        return match
    
    # -----------------------------
    # if not, create new metabolite
    # -----------------------------
    # get information from the database
    bigg_res = load_a_table_from_database(
            f'SELECT * FROM bigg_metabolites WHERE id = \'{id}\'',
            query=True)
    if len(bigg_res) > 0:
        bigg_res = bigg_res.iloc[0,:]
    else:
        return None # not data = no recontruction
    # get information from MNX if ID available
    mnx_res=None
    if bigg_res['MetaNetX (MNX) Chemical']:
        mnx_res = load_a_table_from_database(
            f'SELECT * FROM mnx_chem_prop WHERE id=\'{bigg_res["MetaNetX (MNX) Chemical"]}\'',
            query=True)
        if len(mnx_res) > 0:
            mnx_res = mnx_res.iloc[0,:]
        else:
            mnx_res=None 
    
    # step 1: create a random metabolite ID
    # -------------------------------------
    new_metabolite = cobra.Metabolite(create_random_id(model, 'meta', idprefix)) 
    
    # step 2: add features
    # --------------------
    new_metabolite.name = bigg_res['name']
    new_metabolite.compartment = compartment
    if mnx_res is not None:
        if mnx_res['charge']:
            new_metabolite.charge = mnx_res['charge']
        if mnx_res['formula']:
            new_metabolite.formula = mnx_res['formula']
            
    if not new_metabolite.formula or not new_metabolite.charge:
        try:
            bigg_fetch = json.loads(requests.get(f'http://bigg.ucsd.edu/api/v2/universal/metabolites/{id.rsplit("_",1)[0]}').text)
            if 'formulae' in bigg_fetch.keys() and not new_metabolite.formula:     
                new_metabolite.formula = bigg_fetch['formulae'][0]
            if 'charges' in bigg_fetch.keys() and new_metabolite.charge:   
                new_metabolite.charge = bigg_fetch['charges'][0]
        except Exception as e:
            # @TODO
            pass
    
    # step 3: add notes
    # -----------------
    new_metabolite.notes['created with'] = 'refineGEMs GapFiller, BiGG'

    # step 4: add annotations
    # -----------------------
    # add SBOTerm
    new_metabolite.annotation['sbo'] = 'SBO:0000247'
    # add infos from BiGG
    new_metabolite.annotation['bigg.metabolite'] = [id]  # @TODO or use the universal id?
    add_annotations_from_BiGG_metabs(new_metabolite)
    # add annotations from MNX
    if mnx_res is not None:
        if mnx_res['InChI'] and 'inchi' not in new_metabolite.annotation.keys():
            new_metabolite.annotation['inchi'] = mnx_res['InChI']
        if mnx_res['InChIKey'] and 'inchikey' not in new_metabolite.annotation.keys():
            new_metabolite.annotation['inchikey'] = mnx_res['InChIKey']
    
    # step 5: change ID according to namespace
    # ----------------------------------------
    match_id_to_namespace(new_metabolite,namespace)
    
    # step 6: re-check existence of ID in model
    # -----------------------------------------
    # @TODO : check complete annotations? 
    #        - or let those be covered by the duplicate check later on?
    if new_metabolite.id in [_.id for _ in model.metabolites]:
        return model.metabolites.get_by_id(new_metabolite.id)
    
    return new_metabolite


# adding reactions cobra models
# -----------------------------

def parse_reac_str(equation:str, 
                   type:Literal['BiGG','BioCyc','MetaNetX','KEGG']='MetaNetX') -> tuple[dict,dict,list,bool]:
    """Parse a reaction string.

    Args:
        - equation (str): 
            The equation of a reaction as a string (as saved in the database).
        - type (Literal['BiGG','BioCyc','MetaNetX','KEGG'], optional): 
            The name of the database the equation was taken from. 
            Can be 'BiGG','BioCyc','MetaNetX','KEGG'.
            Defaults to 'MetaNetX'.

    Returns:
        tuple: 
            Tuple of (1) dict, (2) dict, (3) list & (4) bool:
            
            1. Dictionary with the reactant IDs and their stoichiometric factors.
            2. Dictionary with the product IDs and their stoichiometric factors.
            3. List of compartment IDs or None, if they cannot be extract from the equation.
            4. True, if the reaction is reversible, else False.
    """

    products = {}
    reactants = {}
    compartments = list()
    is_product = False
    reversible = True

    match type:
        case 'MetaNetX':
            for s in equation.split(' '):
                # switch from reactants to products
                if s == '=':
                    is_product = True
                # found stoichiometric factor
                elif s.isnumeric():
                    factor = float(s)
                # skip
                elif s == '+':
                    continue
                # found metabolite
                else:
                    # get information from MetaNetX
                    metabolite, compartment = s.split('@')
                    compartments.append(compartment)
                    
                    if is_product:
                        products[metabolite] = factor
                    else:
                        reactants[metabolite] = factor
                        
        case 'BiGG':
            factor = 1.0 # BiGG does not use factor 1 in the equations
            for s in equation.split(' '):
                # found factor
                if s.replace('.','').isdigit():
                    factor = float(s)
                # switch from reactants to products
                elif s == '-->' :
                    is_product = True
                    reversible = False
                elif s == '<->':
                    is_product = True
                # skip
                elif s == '+':
                    continue
                # found metabolite
                else:
                    compartments.append(s.rsplit('_',1)[1])
                    if is_product:
                        products[s] = factor
                    else:
                        reactants[s] = factor
                    factor = 1.0
              
        case 'KEGG':
            compartments = None
            factor = 1.0
            for s in equation.split(' '):
                if s.isnumeric():
                    factor = float(s)
                elif s == '+':
                    continue
                elif s == '<=>': # @TODO are there more options?
                    is_product = True
                else:
                    if is_product:
                        products[s] = factor
                    else:
                        reactants[s] = factor
                    factor = 1.0
                  
    return (reactants,products,compartments,reversible)
        
        
# @TODO
#   extend the build function so, that all of them can take either the id or an equation 
#   as input for rebuilding the reaction (would also be beneficial for semi-manual curation)

@template
# @TODO complete it
def build_reaction_xxx():
    '''
    model:cobra.Model, id:str=None,
                       reac_str:str=None,
                       references:dict={},
                       idprefix:str='refineGEMs',
                       namespace:Literal['BiGG']='BiGG') -> Union[cobra.Reaction, None, list]:
    ''' 
    pass


# @TEST (more) - tries some cases, in which it seems to work
# @TODO: Add if to extend notes if found missing by BioCyc
# @TODO
def build_reaction_mnx(model:cobra.Model, id:str,
                      reac_str:str = None,
                      references:dict={},
                      idprefix:str='refineGEMs',
                      namespace:Literal['BiGG']='BiGG') -> Union[cobra.Reaction, None, list]:
    """Construct a new reaction for a model from a MetaNetX reaction ID.
    This function will NOT add the reaction directly to the model, if the 
    construction process is successful.

    Args:
        - model (cobra.Model): 
            The model loaded with COBRApy.
        - id (str): 
            A MetaNetX reaction ID.
        - reac_str (str, optional): 
            The reaction equation string from the database. 
            Defaults to None.
        - references (dict, optional): 
            Additional annotations to add to the reaction (idtype:[value]). 
            Defaults to {}.
        - idprefix (str, optional): 
            Prefix for the pseudo-identifier. Defaults to 'refineGEMs'.
        - namespace (Literal['BiGG'], optional): 
            Namespace to use for the reaction ID. 
            If namespace cannot be matched, uses the pseudo-ID
            Defaults to 'BiGG'.

    Returns:
        1. Case successful construction:         
                cobra.Reaction:
                    The newly build reaction object. 
            
        2. Case construction not possible:       
                None:
                    Nothing to return.
            
        3. Case reaction found in model.
                list: 
                    List of matching reaction IDs (in model).
    """

    # Get origin if not reaction was found missing with MetaNetX
    origin = references.pop('origin') if 'origin' in references.keys() else None
    
    # ---------------------
    # check, if ID in model
    # ---------------------
    matches_found = [_.id for _ in model.reactions if 'metanetx.reaction' in _.annotation.keys() and _.annotation['metanetx.reaction']==id]
    if len(matches_found) > 0:
        return matches_found
    
    # -----------------------------
    # otherwise, build new reaction
    # -----------------------------
    
    # get relevant part of table from database
    mnx_reac_refs = load_a_table_from_database(
        f'SELECT * FROM mnx_reac_xref WHERE id = \'{id}\'',
        query=True)
    mnx_reac_refs = mnx_reac_refs[~(mnx_reac_refs['description']=='secondary/obsolete/fantasy identifier')]
    
    # create reaction object
    new_reac = cobra.Reaction(create_random_id(model,'reac',idprefix))
    
    # set name of reaction
    name = ''
    for desc in mnx_reac_refs['description']:
        if '|' in desc: # entry has a name and an equation string
            name = desc.split('|')[0]
            break # one name is enough
    new_reac.name = name 
    
    # get metabolites
    # ---------------
    if not reac_str:
        mnx_reac_prop = load_a_table_from_database(
                f'SELECT * FROM mnx_reac_prop WHERE id = \'{id}\'',
                query=True)
        reac_str = mnx_reac_prop['mnx_equation'][0]
        if mnx_reac_prop['ec-code'][0]:
            references['ec-code'] = mnx_reac_prop['ec-code'][0]
        
    if reac_str:
        reactants,products,comparts,rev = parse_reac_str(reac_str,'MetaNetX')
    else:
        return None
    # ............................................................
    # @TODO / Issue
    #    reac_prop / mnx equation only saves generic compartments 1 and 2 (MNXD1 / MNXD2)
    #    how to get the (correct) compartment?
    #    current solution 1 -> c, 2 -> e
    comparts = ['c' if _ == 'MNXD1' else 'e' for _ in comparts ]
    # ............................................................
    metabolites = {}
    meta_counter = 0
    
    # reconstruct reactants
    for mid,factor in reactants.items():
        tmp_meta = build_metabolite_mnx(mid,model,
                                        namespace,
                                        comparts[meta_counter],idprefix)
        if tmp_meta:
            metabolites[tmp_meta] = -1*factor
            meta_counter += 1
        else:
            return None # not able to build reaction successfully
        
    # reconstruct products
    for mid,factor in products.items():
        tmp_meta = build_metabolite_mnx(mid,model,
                                        namespace,
                                        comparts[meta_counter],idprefix)
        if tmp_meta:
            metabolites[tmp_meta] = factor
            meta_counter += 1
        else:
            return None # not able to build reaction successfully
        
    # add metabolites to reaction
    # @TODO: does it need some kind of try and error, if - for some highly unlikely reason - two newly generated ids are the same
    new_reac.add_metabolites(metabolites)
    
    # set reversibility
    if rev:
        new_reac.bounds = cobra.Configuration().bounds
    else:
        new_reac.bounds = (0.0,cobra.Configuration().upper_bound)
        
    # get annotations
    # ---------------
    new_reac.annotation['sbo'] = 'SBO:0000167'
    # get more annotation from the mnx_reac_xref table
    for db in ['bigg.reaction','kegg.reaction','seed.reaction','metacyc.reaction','rhea']:
        db_matches = mnx_reac_refs[mnx_reac_refs['source'].str.contains(db)]
        if len(db_matches) > 0:
            new_reac.annotation[db] = [m.split(':',1)[1] for m in db_matches['source'].tolist()]
            # update reactions direction, if MetaCyc has better information
            if db == 'metacyc.reaction' and len(db_matches[db_matches['source'].str.contains('-->')]):
                new_reac.bounds = (0.0,cobra.Configuration().upper_bound)
    # get alias annotations for MetaNetX if available
    if 'alias' in references.keys(): 
        if references['alias'] != None:
            references['metanetx.reaction'] = list(references['alias'])
        del references['alias']
        
    # add additional references from the parameter
    _add_annotations_from_dict_cobra(references,new_reac)
    
    # get annotations
    # ---------------
    new_reac.annotation['sbo'] = 'SBO:0000167'

    # add notes
    # ---------
    if origin: new_reac.notes['found with'] = f'refineGEMs GapFiller, {origin}'
    new_reac.notes['created with'] = 'refineGEMs GapFiller, MetaNetX'
    
    # match ID to namespace
    # ---------------------
    match_id_to_namespace(new_reac,namespace)
    # re-check, if reaction not in the model based on the namespace
    if new_reac.id in [_.id for _ in model.reactions]:
        return [new_reac.id]
    
    return new_reac


# @TEST (more) - tries some cases, in which it seems to work
# @TODO some things still open (for discussion)
def build_reaction_kegg(model:cobra.Model, id:str=None, reac_str:str=None,
                        references:dict={},
                        idprefix:str='refineGEMs',
                        namespace:Literal['BiGG']='BiGG') -> Union[cobra.Reaction, None, list]:
    """Construct a new reaction for a model from either a KEGG reaction ID
    or a KEGG equation string.
    This function will NOT add the reaction directly to the model, if the 
    construction process is successful.

    Args:
        - model (cobra.Model): 
            The model loaded with COBRApy.
        - id (str,optional): 
            A KEGG reaction ID.
        - reac_str (str, optional): 
            The reaction equation string from the database. 
            Defaults to None.
        - references (dict, optional): 
            Additional annotations to add to the reaction (idtype:[value]). 
            Defaults to {}.
        - idprefix (str, optional): 
            Prefix for the pseudo-identifier. Defaults to 'refineGEMs'.
        - namespace (Literal['BiGG'], optional): 
            Namespace to use for the reaction ID. 
            If namespace cannot be matched, uses the pseudo-ID
            Defaults to 'BiGG'.

    Returns:
        1. Case successful construction:
                cobra.Reaction:
                    The newly build reaction object. 
            
        2. Case construction not possible:
                None:
                    Nothing to return.
            
        3. Case reaction found in model.
                list: 
                    List of matching reaction IDs (in model).
    """
    
    # either reaction id or a reaction string needed for reconstruction
    if not id and not reac_str:
        return None # reconstruction not possible
    
    # create an empty reaction with random id
    new_reac = cobra.Reaction(create_random_id(model,'reac',idprefix))
    
    # -------------
    # KEGG ID given
    # -------------
    if id:
        # check, if reaction in model
        matches = [_.id for _ in model.reactions if 'kegg.reaction' in _.annotation.keys() and _.annotation['kegg.reaction']==id]
        if len(matches) > 0:
            return matches # return matched reaction ids in list
        
        # retrieve information from KEGG
        kegg_info = kegg_reaction_parser(id)
        if kegg_info:
            if 'name' in kegg_info.keys():
                new_reac.name = kegg_info['name']
            if 'equation' in kegg_info.keys():
                reac_str = kegg_info['equation'] if not reac_str else reac_str
            if 'db' in kegg_info.keys():
                new_reac.annotation = kegg_info['db']
    
    # -------------------------------------
    # Reaction reconstruction from equation
    # -------------------------------------
    
    # skip, if no reaction string is available
    if not reac_str:
        return None # reconstruction not possible
    
    # parse reaction string
    # @TODO filter out not useable reaction strings
    reactants,products,comparts,rev = parse_reac_str(reac_str,'KEGG')
        
    # ..............................................
    # @TODO
    # KEGG has no information about compartments !!!
    # current solution: always use c
    compartment = 'c'
    # ..............................................
    metabolites = {}
    meta_counter = 0
    
    # reconstruct reactants
    for mid,factor in reactants.items():
        tmp_meta = build_metabolite_kegg(mid,model,
                                        namespace,
                                        compartment,idprefix)
        if tmp_meta:
            metabolites[tmp_meta] = -1*factor
            meta_counter += 1
        else:
            return None # not able to build reaction successfully
        
    # reconstruct products
    for mid,factor in products.items():
        tmp_meta = build_metabolite_kegg(mid,model,
                                        namespace,
                                        compartment,idprefix)
        if tmp_meta:
            metabolites[tmp_meta] = factor
            meta_counter += 1
        else:
            return None # not able to build reaction successfully
    
    # add metabolites to reaction
    # @TODO: does it need some kind of try and error, if - for some highly unlikely reason - two newly generated ids are the same
    new_reac.add_metabolites(metabolites)
    
    # set reversibility
    if rev:
        new_reac.bounds = (1000.0,1000.0)
    else:
        new_reac.bounds = (0.0,1000.0)
    
    # --------------------
    # add more information
    # --------------------
    new_reac.annotation['sbo'] = 'SBO:0000167'
    # get more information from searching the KEGG ID in BiGG
    bigg_res = load_a_table_from_database(
            f'SELECT * FROM bigg_reactions WHERE \"KEGG Reaction\" = \'{id}\'',
            query=True)
    for idx,row in bigg_res.iterrows():
        r,p,compartments,r = parse_reac_str(row['reaction_string'],'BiGG')
        # .........................................
        # @TODO part 2 of compartment issue
        # find the reaction with 'c' as compartment
        if len(set(compartments)) == 1 and compartments[0] == 'c':
            new_reac.annotation['bigg.reaction'] = row['id']
            # @TODO add more information, exclude None entries
            _add_annotations_from_bigg_reac_row(row, new_reac)
            break
        # .........................................

    # @IDEA / @TODO get more information from MetaNetX
    
    # add additional references from the parameter
    _add_annotations_from_dict_cobra(references,new_reac)
    
    # add notes
    # ---------
    new_reac.notes['created with'] = 'refineGEMs GapFiller, KEGG'
    
    # match ID to namespace
    # ---------------------
    match_id_to_namespace(new_reac,namespace)
    # re-check, if reaction not in the model based on the namespace
    if new_reac.id in [_.id for _ in model.reactions]:
        return [new_reac.id]
    
    return new_reac


# @TEST
# @TODO: Add if to extend notes if found missing by BioCyc
# @TODO some things still open (for discussion)
# @TODO implement reac_str usage
def build_reaction_bigg(model:cobra.Model, id:str, 
                        reac_str:str = None, 
                        references:dict={},
                        idprefix:str='refineGEMs',
                        namespace:Literal['BiGG']='BiGG') -> Union[cobra.Reaction, None, list]:
    """Construct a new reaction for a model from a BiGG reaction ID.
    This function will NOT add the reaction directly to the model, if the 
    construction process is successful.

    Args:
        - model (cobra.Model): 
            The model loaded with COBRApy.
        - id (str): 
            A BiGG reaction ID.
        - reac_str (str, optional): 
            The reaction equation string from the database. 
            Currently, this param is not doing anything in this function.
            Defaults to None.
        - references (dict, optional): 
            Additional annotations to add to the reaction (idtype:[value]). 
            Defaults to {}.
        - idprefix (str, optional): 
            Prefix for the pseudo-identifier. Defaults to 'refineGEMs'.
        - namespace (Literal['BiGG'], optional): 
            Namespace to use for the reaction ID. 
            If namespace cannot be matched, uses the pseudo-ID
            Defaults to 'BiGG'.

    Returns:
        1. Case successful construction:
                cobra.Reaction:
                    The newly build reaction object. 
            
        2. Case construction not possible:
                None:
                    Nothing to return.
            
        3. Case reaction found in model.
                list: 
                    List of matching reaction IDs (in model).
    """

    # Get origin if not reaction was found missing with MetaNetX
    origin = references.pop('origin') if 'origin' in references.keys() else None
    
    # ---------------------
    # check, if ID in model
    # ---------------------
    matches_found = [_.id for _ in model.reactions if 'bigg.reaction' in _.annotation.keys() and _.annotation['bigg.reaction']==id]
    if len(matches_found) > 0:
        return matches_found
    
    # -----------------------------
    # otherwise, build new reaction
    # -----------------------------
    # create reaction object
    new_reac = cobra.Reaction(create_random_id(model,'reac',idprefix))
    
    # get information from the database
    bigg_reac_info = load_a_table_from_database(
            f'SELECT * FROM bigg_reactions WHERE id = \'{id}\'',
            query=True).iloc[0,:]
    new_reac.name = bigg_reac_info['name']
    
    # add metabolites
    # ---------------
    reactants,products,comparts,rev = parse_reac_str(bigg_reac_info['reaction_string'],'BiGG')
    
    metabolites = {}
    meta_counter = 0
    # reconstruct reactants
    for mid,factor in reactants.items():
        tmp_meta = build_metabolite_bigg(mid,model,
                                        namespace,
                                        idprefix)
        if tmp_meta:
            metabolites[tmp_meta] = -1*factor
            meta_counter += 1
        else:
            return None # not able to build reaction successfully
        
    # reconstruct products
    for mid,factor in products.items():
        tmp_meta = build_metabolite_bigg(mid,model,
                                        namespace,
                                        idprefix)
        if tmp_meta:
            metabolites[tmp_meta] = factor
            meta_counter += 1
        else:
            return None # not able to build reaction successfully
        
    # add metabolites to reaction
    # @TODO: does it need some kind of try and error, if - for some highly unlikely reason - two newly generated ids are the same
    new_reac.add_metabolites(metabolites)
    
    # set reversibility
    if rev:
        new_reac.bounds = (1000.0,1000.0)
    else:
        new_reac.bounds = (0.0,1000.0)
    
    # add annotations
    # ---------------
    # add SBOTerm
    new_reac.annotation['sbo'] = 'SBO:0000167'
    # add infos from BiGG
    new_reac.annotation['bigg.reaction'] = [id]
    _add_annotations_from_bigg_reac_row(bigg_reac_info, new_reac)
    # get alias annotations for MetaNetX if available
    if 'alias' in references.keys(): 
        if references['alias'] != None:
            references['metanetx.reaction'] = list(references['alias'])
        del references['alias']
    # add additional references from the parameter
    _add_annotations_from_dict_cobra(references,new_reac)
    
    # add notes
    # ---------
    if origin: new_reac.notes['found with'] = f'refineGEMs GapFiller, {origin}'
    new_reac.notes['created with'] = 'refineGEMs GapFiller, BiGG'
    
    # match ID to namespace
    # ---------------------
    match_id_to_namespace(new_reac,namespace)
    # re-check, if reaction not in the model based on the namespace
    if new_reac.id in [_.id for _ in model.reactions]:
        return [new_reac.id]

    return new_reac


@implement
# maybe for later, if we need something to work independantly from namespace
# and outside the gapfilling
# Or would it be better to have a function add_reaction() that only adds 
# reactions built with the previuous functions?
def build_reaction():
    pass


# ++++++++++++++++++++++++++++++++++++++++
# libSBML - models
# ++++++++++++++++++++++++++++++++++++++++

# extracting reactions & Co via libsbml
# -------------------------------------

# Function originally from refineGEMs.genecomp/refineGEMs.KEGG_analysis --- Modified
# Might be possible to deprecate
def compare_gene_lists(gps_in_model: pd.DataFrame, db_genes: pd.DataFrame, kegg: bool=True) -> pd.DataFrame:
    """Compares the provided tables according to column 0/'Locus_tag'
    
    Args:
        - gps_in_model (pd.DataFrame): 
            Table containing the KEGG Gene IDs/Locus tags in the model
        - db_genes (pd.DataFrame): 
            Table containing the KEGG Gene IDs for the organism from KEGG/
            locus tags (Accession-2) from BioCyc
        - kegg (bool): 
            True if KEGG Genes should be extracted, otherwise False
        
    Returns:
        pd.DataFrame: 
            Table containing all missing genes
    """
    in_db = db_genes.set_index(0) if kegg else db_genes.set_index('locus_tag')
    in_model = gps_in_model.set_index(0)
        
    genes_in_db_not_in_model = in_db[~in_db.index.isin(in_model.index)]
    
    return genes_in_db_not_in_model.reset_index().iloc[:, 0] if kegg else genes_in_db_not_in_model.reset_index()


# Function originally from refineGEMs.genecomp/refineGEMs.KEGG_analysis --- Modified
def get_model_reacs_or_metabs(model_libsbml: libModel, metabolites: bool=False, col_name: str='bigg_id') -> pd.DataFrame:
    """Extracts table of reactions/metabolites with BiGG IDs from model

    Args:
        - model_libsbml (libModel): 
            Model loaded with libSBML
        - metabolites (bool): 
            Set to True if metabolites from model should be extracted
        - col_name (str): 
            Name to be used for column in table. Defaults to 'bigg_id'.

    Returns:
        pd.DataFrame: 
            Table with model identifiers for either metabolites or reactions
    """
    reac_or_metab_list = model_libsbml.getListOfSpecies() if metabolites else model_libsbml.getListOfReactions()

    list_of_reacs_or_metabs = []
    for reac_or_metab in reac_or_metab_list:
        list_of_reacs_or_metabs.append(reac_or_metab.id[2:])

    reac_or_metab_list_df = pd.Series(list_of_reacs_or_metabs)
    reac_or_metab_list_df = pd.DataFrame(reac_or_metab_list_df, columns=[col_name])

    return reac_or_metab_list_df


def get_reversible(fluxes: dict[str: str]) -> bool:
    """Infer if reaction is reversible from flux bounds
    
    Args:
        - fluxes (dict): 
            Dictionary containing the keys 'lower_bound' & 'upper_bound' 
            with values in ['cobra_default_lb', 'cobra_0_bound', 'cobra_default_ub']
    
    Returns:
        bool: 
            True if reversible else False
    """
    return (fluxes['lower_bound'] == 'cobra_default_lb') and (fluxes['upper_bound'] == 'cobra_default_ub')


# create model entities using libSBML
# -----------------------------------

# @TODO check for new-ish functionalities / merge with create_gp and delete
def create_gpr_from_locus_tag(model: libModel, locus_tag: str, email: str) -> tuple[GeneProduct, libModel]:
    """Creates GeneProduct in the given model

    **Deprecation warning**: will be deprecated in a future update.

    Args:
        - model (libModel): 
            Model loaded with libSBML
        - locus_tag (str): 
            NCBI compatible locus_tag
        - email (str): 
            User Email to access the NCBI Entrez database

    Returns:
        tuple: 
            libSBML GeneProduct (1) & libSBML model (2)

            (1) GeneProduct: Created gene product
            (2) libModel: Model containing the created gene product
    """
    mes = f'create_gpr_from_locus_tag will be deprecated in a future update.'
    warnings.warn(mes,type=FutureWarning)

    Entrez.email = email
    name, locus = search_ncbi_for_gpr(locus_tag)
    gpr = model.getPlugin(0).createGeneProduct()
    gpr.setName(name)
    gpr.setId(locus_tag)
    gpr.setMetaId('meta_' + locus_tag)
    gpr.setLabel(locus_tag)
    gpr.setSBOTerm("SBO:0000243")
    add_cv_term_genes(locus_tag, 'NCBI', gpr)
    return gpr, model


# @TODO : check, if the function (in ths way), is still used anywhere and adjust to the 
# new one
# @DEPRECATE
def old_create_gp(model: libModel, model_id: str, name: str, locus_tag: str, protein_id: str) -> tuple[GeneProduct, libModel]:
    """Creates GeneProduct in the given model

    Args:
        - model (libModel): 
            Model loaded with libSBML
        - model_id (str): 
            ID identical to ID that CarveMe adds from the NCBI FASTA input file
        - name (str): 
            Name of the GeneProduct
        - locus_tag (str): 
            Genome-specific locus tag used as label in the model
        - protein_id (str): 
            NCBI Protein/RefSeq ID

    Returns:
        tuple: 
            libSBML GeneProduct (1) & libSBML model (2)

            (1) GeneProduct: Created gene product
            (2) libModel: Model containing the created gene product
    """
    id_db = None
    gp = model.getPlugin(0).createGeneProduct()
    gp.setId(model_id) # libsbml advised to use set/getIdAttribute
    gp.setName(name)
    gp.setLabel(locus_tag)
    gp.setSBOTerm('SBO:0000243')
    gp.setMetaId(f'meta_{model_id}')
    if re.fullmatch('^(((AC|AP|NC|NG|NM|NP|NR|NT|NW|WP|XM|XP|XR|YP|ZP)_\d+)|(NZ_[A-Z]{2,4}\d+))(\.\d+)?$', protein_id, re.IGNORECASE):
        id_db = 'REFSEQ'
    elif re.fullmatch('^(\w+\d+(\.\d+)?)|(NP_\d+)$', protein_id, re.IGNORECASE): id_db = 'NCBI'
    if id_db: add_cv_term_genes(protein_id, id_db, gp)
    return gp, model


# @NEW - substitues the one above 
# --> WARNING: Output is different now
# @TODO generalise addition of references -> maybe kwargs
# @TODO: Check if ncbiprotein leads to valid ID -> Otherwise, replace invalid chars with '_'
def create_gp(model:libModel, protein_id:str, 
              model_id:str=None,
              name:str=None, locus_tag:str=None,
              uniprot:tuple[list,bool]=None) -> None:
    """Creates GeneProduct in the given libSBML model.

    Args:
        - model (libModel): 
            The model object, loaded with libSBML.
        - protein_id (str): 
            (NCBI) Protein ID of the gene.
        - model_id (str, optional): 
            If given, uses this string as the ID of the gene in the model. 
            ID should be identical to ID that CarveMe adds from the NCBI FASTA input file.
            Defaults to None.
        - name (str, optional): 
            Name of the GeneProduct. Defaults to None.
        - locus_tag (str, optional): 
            Genome-specific locus tag. Will be used as label in the model. 
            Defaults to None.
        - uniprot (tuple[list,bool], optional): 
            Tuple of a list of UniProt IDs and a boolean for whether the strain is from the lab or
            a database. Defaults to None.
    """
    
    # create gene product object
    gp = model.getPlugin(0).createGeneProduct()
    # set basic attributes
    if model_id:                            # ID 
        gp.setIdAttribute(model_id)
    else:
        geneid = f'G_{protein_id}'.replace('.','_') # remove problematic signs
        gp.setIdAttribute(geneid)               
    if name: gp.setName(name)               # Name  
    if locus_tag: gp.setLabel(locus_tag)    # Label
    gp.setSBOTerm('SBO:0000243')            # SBOterm
    gp.setMetaId(f'meta_G_{protein_id}')    # Meta ID
    # test for NCBI/RefSeq
    id_db = None
    if re.fullmatch('^(((AC|AP|NC|NG|NM|NP|NR|NT|NW|WP|XM|XP|XR|YP|ZP)_\d+)|(NZ_[A-Z]{2,4}\d+))(\.\d+)?$', protein_id, re.IGNORECASE):
        id_db = 'REFSEQ'
    elif re.fullmatch('^(\w+\d+(\.\d+)?)|(NP_\d+)$', protein_id, re.IGNORECASE): id_db = 'NCBI'
    if id_db: add_cv_term_genes(protein_id, id_db, gp)           # NCBI protein
    # add further references
    # @TODO extend or generalise
    if uniprot:
        for uniprotid in uniprot[0]:
            add_cv_term_genes(uniprotid, 'UNIPROT', gp, uniprot[1]) # UniProt
      

def create_species(
    model: libModel, metabolite_id: str, name: str, compartment_id: str, charge: int, chem_formula: str
                   ) -> tuple[Species, libModel]:
    """Creates Species/Metabolite in the given model

    Args:
        - model (libModel): 
            Model loaded with libSBML
        - metabolite_id (str): 
            Metabolite ID within model (If model from CarveMe, preferable a BiGG ID)
        - name (str): 
            Name of the metabolite
        - compartment_id (str): 
            ID of the compartment where metabolite resides
        - charge (int): 
            Charge for the metabolite
        - chem_formula (str): 
            Chemical formula for the metabolite

    Returns:
        tuple: 
            libSBML Species (1) & libSBML model (2)
            
            (1) Species: Created species/metabolite
            (2) libModel: Model containing the created metabolite
    """
    metabolite = model.createSpecies()
    metabolite.setId(f'M_{metabolite_id}')
    if name: metabolite.setName(name)
    metabolite.setMetaId(f'meta_M_{metabolite_id}')
    metabolite.setSBOTerm('SBO:0000247')
    metabolite.setInitialAmount(float('NaN'))
    metabolite.setHasOnlySubstanceUnits(True)
    metabolite.setBoundaryCondition(False)
    metabolite.setConstant(False)
    metabolite.setCompartment(compartment_id)
    metabolite.getPlugin(0).setCharge(charge)
    metabolite.getPlugin(0).setChemicalFormula(chem_formula)
    add_cv_term_metabolites(metabolite_id[:-2], 'BIGG', metabolite)
    return metabolite, model


def create_reaction(
    model: libModel, reaction_id: str, name:str, reactants: dict[str: int], products: dict[str: int], 
    fluxes: dict[str: str], reversible: bool=None, fast: bool=None, compartment: str=None, sbo: str=None, 
    genes: Union[str, list[str]]=None
    ) -> tuple[Reaction, libModel]:
    """Creates new reaction in the given model

    Args:
        - model (libModel): 
            Model loaded with libSBML
        - reaction_id (str): 
            BiGG ID of the reaction to create
        - name (str): 
            Human readable name of the reaction
        - reactants (dict): 
            Metabolites as keys and their stoichiometry as values
        - products (dict): 
            Metabolites as keys and their stoichiometry as values
        - fluxes (dict): 
            Dictionary with lower_bound and upper_bound as keys
        - reversible (bool): 
            True/False for the reaction
        - fast (bool): 
            True/False for the reaction
        - compartment (str): 
            BiGG compartment ID of the reaction (if available)
        - sbo (str): 
            SBO term of the reaction
        - genes (str|list): 
            List of genes belonging to reaction

    Returns:
        tuple: 
            libSBML reaction (1) & libSBML model (2)
        
            (1) Reaction: Created reaction 
            (2) libModel: Model containing the created reaction
    """
    reaction = model.createReaction()
    reaction.setId('R_' + reaction_id)
    if name: reaction.setName(name)
    reaction.setMetaId('meta_R_' + reaction_id)
    sbo = sbo if sbo else 'SBO:0000167'  # SBO term for biochemical or transport reaction
    reaction.setSBOTerm(sbo)
    fast = fast if fast else False
    reaction.setFast(fast)
    if compartment: reaction.setCompartment(compartment)  # Set compartment for reaction if available
    reversible = reversible if reversible else get_reversible(fluxes)
    reaction.setReversible(reversible)
    if genes:
            if genes == 'G_spontaneous':
                reaction.getPlugin(0).createGeneProductAssociation().createGeneProductRef().setGeneProduct(gene)
            elif len(genes) == 1:
                reaction.getPlugin(0).createGeneProductAssociation().createGeneProductRef().setGeneProduct(genes[0])
            else:
                gp_ass_or = reaction.getPlugin(0).createGeneProductAssociation().createOr()
                for gene in genes:
                    # Set GeneProductReferences if available
                    gp_ass_or.createGeneProductRef().setGeneProduct(gene)
    for metab, stoich in reactants.items(): #reactants as dict with metab:stoich
        reaction.addReactant(model.getSpecies('M_' + metab), stoich)
    for metab, stoich in products.items(): #reactants as dict with metab:stoich
        reaction.addProduct(model.getSpecies('M_' + metab), stoich)
    reaction.getPlugin(0).setLowerFluxBound(fluxes['lower_bound'])
    reaction.getPlugin(0).setUpperFluxBound(fluxes['upper_bound'])
    add_cv_term_reactions(reaction_id, 'BIGG', reaction)
    return reaction, model
 

# @TODO : does it cover indeed all cases (for adding GPR together) ?
# @TODO : only support OR connection
def create_gpr(reaction:Reaction,gene:str|list[str]) -> None:
    """For a given libSBML Reaction and a gene ID or a list of gene IDs, 
    create a gene production rule inside the reaction.
    
    Currently only supports 'OR' causality.

    Args:
        - reaction (libsbml.Reaction): 
            The reaction object to add the GPR to.
        - gene (str | list[str]): 
            Either a gene ID or a list of gene IDs, that will be added to the GPR 
            (OR causality).
    """

    # Step 1: test, if there is already a gpr
    # ---------------------------------------
    old_association_str = None
    old_association_fbc = None
    if reaction.getPlugin(0).getGeneProductAssociation():
        old_association = reaction.getPlugin(0).getGeneProductAssociation().getListOfAllElements()
        # case 1: only a single association
        if len(old_association) == 1 and isinstance(old_association[0],GeneProductRef):
            old_association_str = old_association[0].getGeneProduct()
        # case 2: nested structure of asociations
        elif isinstance(old_association[0], FbcOr) or isinstance(old_association[0], FbcAnd):
            old_association_fbc = old_association[0].clone()
            # this should get the highest level association (that includes all others)

                    
    # Step 2: create new gene product association 
    # -------------------------------------------

    if old_association_str and isinstance(gene,str):
        gene = [old_association_str,gene]
    elif old_association_str  and isinstance(gene,list):
        gene.append(old_association_str)
        
    # add the old association rule as an 'OR' (if needed)
    if not old_association_fbc:
        new_association = reaction.getPlugin(0).createGeneProductAssociation()
    else:
        new_association = reaction.getPlugin(0).createGeneProductAssociation().createOr()
        new_association.addAssociation(old_association_fbc)

    # add the remaining genes 
    # @TODO currently, only connection possible is 'OR'
    if isinstance(gene,str):
        new_association.createGeneProductRef().setGeneProduct(gene)
    elif isinstance(gene,list) and len(gene) == 1:
        new_association.createGeneProductRef().setGeneProduct(gene[0])
    elif isinstance(gene,list) and len(gene) > 1:
        gpa_or =  new_association.createOr()
        for i in gene:
            gpa_or.createGeneProductRef().setGeneProduct(i)

            



__all__ = ['connections', 'cvterms', 'databases', 'db_access', 'entities','io','set_up','util']

from . import connections
from . import cvterms
from . import databases
from . import db_access
from . import entities
from . import io
from . import set_up
from . import util


#!/usr/bin/env python
""" Provides functions to investigate the model and test with MEMOTE

These functions enable simple testing of any model using MEMOTE and access to its number of reactions, metabolites and genes.
"""

__author__ = "Famke Baeuerle and Alina Renz and Carolin Brune"

################################################################################
# requirements
################################################################################

import pandas as pd
import cobra

from libsbml import Model as libModel
from cobra import Model as cobraModel

from memote.support import consistency
# needed by memote.support.consistency
from memote.support import consistency_helpers as con_helpers

from ..utility.io import search_sbo_label
from ..utility.util import test_biomass_presence

################################################################################
# variables
################################################################################

################################################################################
# functions
################################################################################

# get basic model info
# --------------------

def get_reac_with_gpr(model:cobra.Model) -> tuple[list[str], list[str]]:
    """Extract the reactions that have a gene production rule from a given model.

    Args:
        - model (cobra.Model): 
            The model loaded with COBRApy.

    Returns:
        tuple: 
            Lists of reactions with GPR (1) & (2):

            (1) list: List of normal reactions with gpr
            (2) list: List of pseudoreactions with gpr
    """
    pseudoreactions = model.boundary
    for biomass in test_biomass_presence(model):
        pseudoreactions.append(model.reactions.get_by_id(biomass))

    normal = []
    pseudo = []
    for reac in model.reactions:
        # check for GPR
        if len(reac.genes) > 0:
            if reac in pseudoreactions:
                pseudo.append(reac.id)
            else:
                normal.append(reac.id)

    return (normal, pseudo)


def get_orphans_deadends_disconnected(model: cobraModel) -> tuple[list[str], list[str], list[str]]:
    """Uses MEMOTE functions to extract orphans, deadends and disconnected metabolites

    Args:
        - model (cobraModel): 
            Model loaded with COBRApy

    Returns:
        tuple: 
            Lists of metabolites that might cause errors (1) - (3) 

            (1) list: List of orphans
            (2) list: List of deadends
            (3) list: List of disconnected metabolites
    """
    orphans = consistency.find_orphans(model)
    deadends = consistency.find_deadends(model)
    disconnected = consistency.find_disconnected(model)

    orphan_list = []
    if len(orphans) > 0:
        for orphan in orphans:
            orphan_list.append(orphan.id)

    deadend_list = []
    if len(deadends) > 0:
        for deadend in deadends:
            deadend_list.append(deadend.id)

    disconnected_list = []
    if len(disconnected) > 0:
        for disc in disconnected:
            disconnected_list.append(disc.id)

    return orphan_list, deadend_list, disconnected_list


# @TODO: what about exchange reactions not starting with an EX - as usually the case within the BiGG namespace? 
def get_mass_charge_unbalanced(model: cobraModel) -> tuple[list[str], list[str]]:
    """Creates lists of mass and charge unbalanced reactions, without exchange reactions since they are unbalanced per definition

    Args:
        - model (cobraModel): 
            Model loaded with COBRApy

    Returns:
        tuple: 
            Lists of reactions that might cause errors (1) & (2)
            
            (1) list: List of mass unbalanced reactions
            (2) list: List of charge unbalanced reactions
    """

    mass_unbalanced = consistency.find_mass_unbalanced_reactions(
        model.reactions)
    charge_unbalanced = consistency.find_charge_unbalanced_reactions(
        model.reactions)

    mass_list = []
    if len(mass_unbalanced) > 0:
        for reac in mass_unbalanced:
            if (reac.id[:2] != 'EX'):
                mass_list.append(reac.id)

    charge_list = []
    if len(charge_unbalanced) > 0:
        for reac in charge_unbalanced:
            if (reac.id[:2] != 'EX'):
                charge_list.append(reac.id)

    return mass_list, charge_list


# other
# -----

def get_metabs_with_one_cvterm(model: libModel) -> list[str]:
    """Reports metabolites which have only one annotation, can be used as basis for further annotation research

    Args:
        - model (libModel): 
            Model loaded with libSBML

    Returns:
        list: 
            Metabolite Ids with only one annotation
    """
    spe = model.getListOfSpecies()

    only_one = [] #safe metab with only BiGG annotation
    for sb in spe:
        if sb.isSetId():
            pid = sb.getId()
            if sb.getCVTerm(0).getNumResources() == 1:
                only_one.append(pid)
                
    return only_one

def get_reactions_per_sbo(model: libModel) -> dict:
    """Counts number of reactions of all SBO Terms present

    Args:
        - model (libModel): 
            Model loaded with libSBML

    Returns:
        dict: 
            SBO Term as keys and number of reactions as values
    """
    sbos_dict = {}
    for react in model.getListOfReactions():
        sbo = react.getSBOTerm()
        if sbo in sbos_dict.keys():
            sbos_dict[sbo] += 1
        else: 
            sbos_dict[sbo] = 1
    return sbos_dict

def plot_rea_sbo_single(model: libModel):
    """Plots reactions per SBO Term in horizontal bar chart

    Args:
        - model (libModel): 
            Model loaded with libSBML

    Returns:
        plot: 
            Pandas Barchart
    """
    df = pd.DataFrame(get_reactions_per_sbo(model), index=[0]).T.reset_index().rename({0:model.id, 'index': 'SBO-Term'}, axis=1)
    df = df[df[model.id]>3]
    df['SBO-Name'] = df['SBO-Term'].apply(search_sbo_label)
    ax = df.drop('SBO-Term', axis=1).sort_values(model.id).set_index('SBO-Name').plot.barh(width=.8, figsize=(8,10))
    ax.set_ylabel('')
    ax.set_xlabel('number of reactions', fontsize=16)
    ax.legend(loc='lower right')
    fig = ax.get_figure()
    return fig


#!/usr/bin/env python
"""General functions for curating a model

Includes functions for adding annotations from manual tables, annotations synchronisation, 
handling duplicates and more."""

__author__ = "Famke Baeuerle und Carolin Brune"

################################################################################
# requirements
################################################################################

import cobra
import pandas as pd
import re
import warnings

from libsbml import Model as libModel
from tqdm.auto import tqdm
from typing import Literal

from ..utility.cvterms import add_cv_term_metabolites, metabol_db_dict, get_id_from_cv_term
from ..utility.util import test_biomass_presence

################################################################################
# variables
################################################################################

NH_PATTERN = re.compile('nh[3-4]') #: :meta: 

################################################################################
# functions
################################################################################
    
# sychronise annotations
# ----------------------

def update_annotations_from_others(model: libModel) -> libModel:
    """Synchronizes metabolite annotations for core, periplasm and extracelullar

    Args:
        - model (libModel): 
            Model loaded with libSBML

    Returns:
        libModel: 
            Modified model with synchronized annotations
    """
    for metab in model.getListOfSpecies():
        base = metab.getId()[:-2]
        for comp in ['_c', '_e', '_p']:
            other_metab = model.getSpecies(base + comp)
            if other_metab is not None:
                if not other_metab.isSetMetaId():
                    other_metab.setMetaId('meta_' + other_metab.getId())
                for db_id, code in metabol_db_dict.items():
                    id = get_id_from_cv_term(metab, code)
                    for entry in id:
                        if entry is not None:
                            add_cv_term_metabolites(entry, db_id, other_metab)    
    return model               


def complete_BioMetaCyc(model:cobra.Model) -> cobra.Model:
    """Check for existing MetaCyc and BioCyc annotations for metabolites and
    reactions and generate them from the other if one of the two is missing.

    Args:
        model (cobra.Model): 
            A genome-scale model to be checked 
            for complete BioCyc/MetaCyc annotations.

    Returns:
        cobra.Model: 
            The updated model.
    """


    # reactions
    # ---------
    for reac in model.reactions:
        # case 1: MetaCyc but not BioCyc
        if 'metacyc.reaction' in reac.annotation and not 'biocyc' in reac.annotation:
            # add database (META) information to get BioCyc ID
            if isinstance(reac.annotation['metacyc.reaction'], list):
                reac.annotation['biocyc'] = ['META:' + _ for _ in reac.annotation['metacyc.reaction']]
            else:
                reac.annotation['biocyc'] = 'META:' + reac.annotation['metacyc.reaction']
        # case 2: BioCyc, but no MetaCyc
        elif 'biocyc' in reac.annotation and not 'metacyc.reaction' in reac.annotation:
            # if there are multiple metacyc.reaction annotation
            if isinstance(reac.annotation['biocyc'], list):
                add_anno = []
                for biocyc_anno in reac.annotation['biocyc']:
                    if ':' in biocyc_anno:
                        # exclude organism identifier from ID to get MetaCyc ID
                        add_anno.append(biocyc_anno.split(':')[1])
                    else:
                        # high possibility that information is faulty - do not use it
                        print(F'\n\nWarning: Unexpected BioCyc annotation {biocyc_anno} for reaction {reac.id}')
                reac.annotation['metacyc.reaction'] = add_anno
            # if there is only one
            else:
                if ':' in reac.annotation['biocyc']:
                    # exclude organism identifier from ID to get MetaCyc ID
                    reac.annotation['metacyc.reaction'] = reac.annotation['biocyc'].split(':')[1]
                else:
                    # high possibility that information is faulty - do not use it
                    print(F'\n\nWarning: Unexpected BioCyc annotation {reac.annotation["biocyc"]} for reaction {reac.id}')
        # case 3: both or no = skip
        else:
            continue

    # metabolites
    # -----------
    for meta in model.metabolites:
        # case 1: MetaCyc but not BioCyc
        if 'metacyc.compound' in meta.annotation and not 'biocyc' in meta.annotation:
            # add database (META) information to get BioCyc ID
            if isinstance(meta.annotation['metacyc.compound'],list):
                meta.annotation['biocyc'] = ['META:' + _ for _ in meta.annotation['metacyc.compound']]
            else:
                meta.annotation['biocyc'] = 'META:' + meta.annotation['metacyc.compound']
        # case 2: BioCyc, but no MetaCyc
        elif 'biocyc' in meta.annotation and not 'metacyc.compound' in meta.annotation:
            # if there are multiple metacyc.compound annotations
            if isinstance(meta.annotation['biocyc'], list):
                add_anno = []
                for biocyc_anno in meta.annotation['biocyc']:
                    if ':' in biocyc_anno:
                        # exclude organism identifier from ID to get MetaCyc ID
                        add_anno.append(biocyc_anno.split(':')[1])
                    else:
                        # high possibility that information is faulty - do not use it
                        print(F'\n\nWarning: Unexpected BioCyc annotation {biocyc_anno} for metabolite {meta.id}')
                meta.annotation['metacyc.compound'] = add_anno
            # if there is only one
            else:
                if ':' in meta.annotation['biocyc']:
                    # exclude organism identifier from ID to get MetaCyc ID
                    meta.annotation['metacyc.compound'] = meta.annotation['biocyc'].split(':')[1]
                else:
                    # high possibility that information is faulty - do not use it
                    print(F'\n\nWarning: Unexpected BioCyc annotation {meta.annotation["biocyc"]} for metabolite {meta.id}')
        # case 3: both or no = skip
        else:
            continue

    return model



# duplicates
# ----------

def resolve_duplicate_reactions(model:cobra.Model, based_on:str='reaction', remove_reac:bool=True) -> cobra.Model:
    """Resolve and remove duplicate reaction based on their reaction equation
    and matching database identifiers. Only if all match or a comparison with nan occurs will one of
    the reactions be removed.

    Args:
        - model (cobra.Model): 
            A model loaded with COBRApy.
        - based_on (str, optional): 
            Label to base the resolvement process on . 
            Can be 'reaction' or any other annotation label. 
            Defaults to 'reaction'.
        - remove_reac (bool, optional): 
            When True, combines and remove duplicates. 
            Otherwise only reports the findings.
            Defaults to True.

    Returns:
        cobra.Model: 
            The model.
    """

    # get annotation and compartment information
    anno_reac = []
    for r in model.reactions:
        anno_reac.append({'id':r.id, 'compartment':str(r.compartments), 'reaction':r.reaction} | r.annotation)
    df_reac = pd.DataFrame.from_dict(anno_reac)

    # check if based_on is valid
    if not based_on in df_reac.columns.tolist():
        warnings.warn(F'Warning: Annotation column {based_on} does not exists. Search for duplicates will be skipped.')
        return model

    # set basic parameters
    skip_cols = ['id','compartment','bigg.reaction','reaction',based_on]
    colnames = df_reac.columns.tolist()

    for c in df_reac.groupby('compartment'):
        # note: using groupby drops nans
        for mnx in c[1].groupby(based_on):
            # find possible duplicates
            dupl = True
            annotations = {}
            if len(mnx[1]) > 1:
                # check annotations
                for col in [_ for _ in colnames if not _ in skip_cols]:
                    if len(mnx[1][col].dropna().value_counts()) < 2:
                        annotations[col] = mnx[1][col].dropna().explode().unique().tolist()
                    else:
                        dupl = False
                        break

                # if duplicate found
                if dupl:
                    if remove_reac:

                        # choose reaction to keep
                        keep_reac = model.reactions.get_by_id(mnx[1]['id'].tolist()[0])

                        # resolve annotations
                        for key, value in annotations.items():
                            if len(value) > 0 and not key in keep_reac.annotation:
                                keep_reac.annotation[key] = value

                        # combine gene reaction rules
                        for r_id in mnx[1]['id'].tolist()[1:]:
                            gpr_to_add = model.reactions.get_by_id(r_id).gene_reaction_rule
                            if gpr_to_add and gpr_to_add != '':
                                if keep_reac.gene_reaction_rule and  keep_reac.gene_reaction_rule != '': # add two existing rules
                                     keep_reac.gene_reaction_rule =  keep_reac.gene_reaction_rule + ' or ' + gpr_to_add
                                else:
                                    keep_reac.gene_reaction_rule = gpr_to_add # add the one existing the other 
                            else:
                                pass # nothing to add
                            model.reactions.get_by_id(r_id).remove_from_model() # @NOTE throws warning, but seems to be a cobra-internal issue
                            print(F'\tDuplicate reaction {r_id} found. Combined to {keep_reac.id} and deleted.')
                    else:
                        print(f'\tDuplicate reactions {", ".join(mnx[1]["id"].tolist())} found.')

    return model


def resolve_duplicate_metabolites(model:cobra.Model, based_on:str='metanetx.chemical', replace:bool=True) -> cobra.Model:
    """Resolve duplicate metabolites in a model. Metabolites are considered
    duplicate if they share the same annotations (same or nan).
    
    .. note:: 
        Depending on the starting database, the results might differ.

    Args:
        - model (cobra.Model): 
            The model loaded with COBRApy.
        - based_on (str, optional): 
            Label to base the resolvement process on . 
            Can be any annotation label. 
            Defaults to 'metanetx.chemical'.
        - replace (bool, optional): 
            Either report the duplicates (False) 
            or replace them with one (True). Defaults to True.

    Returns:
        cobra.Model: 
            The model.
    """

    # get annotation and compartment information
    anno_meta = []
    for m in model.metabolites:
        anno_meta.append({'id':m.id, 'compartment':m.compartment} | m.annotation)
    df_meta = pd.DataFrame.from_dict(anno_meta)

    # check if based_on is valid
    if not based_on in df_meta.columns.tolist():
        warnings.warn(F'Warning: Annotation {based_on} not found. Search for metabolite duplicates skipped.')
        return model

    # set basic parameters
    skip_cols = ['id','compartment','bigg.metabolite',based_on]
    colnames = df_meta.columns.tolist()
    # get objective function
    bof_list = test_biomass_presence(model)
    if len(bof_list) == 1:
        objective_function = bof_list[0]
    elif len(bof_list) > 1:
        mes = f'Multiple BOFs detected. Will be using {bof_list[0]}'
        warnings.warn(mes,category=UserWarning)
        objective_function = bof_list[0]
    else:
        mes = 'No BOF detected. Might lead to problems during duplicate removal.'
        warnings.warn(mes, category=UserWarning) 

    for c in df_meta.groupby('compartment'):
        # note: using groupby drops nans
        # note: mnx as starting point was choosen, as it seems to have the most annotations (easy to get)
        for mnx in c[1].groupby(based_on):

            # find possible duplicates
            dupl = True
            annotations = {}
            if len(mnx[1]) > 1:
                for col in [_ for _ in colnames if not _ in skip_cols]:
                    # check annotation, if current group truly consists of duplicates
                    if len(mnx[1][col].dropna().value_counts()) < 2:
                        annotations[col] = mnx[1][col].dropna().explode().unique().tolist()
                    else:
                        dupl = False
                        break
            # no duplicates = check next
            else:
                continue

            # if duplicate is found:
            if dupl:
                # either replace ...
                if replace:
                    # choose metabolite to keep
                    keep_meta = model.metabolites.get_by_id(mnx[1]['id'].tolist()[0])
                    # resolve annotations
                    for key, value in annotations.items():
                        if len(value) > 0 and not key in keep_meta.annotation:
                            keep_meta.annotations[key] = value
                    # note: charge and formula should be in valid range and be corrected by MCC (if needed)

                    # replace duplicates with the metabolite to be kept
                    #     to ensure consistency, only delete duplicate metabolites, which
                    #     do NOT share ANY reactions

                    # retrieve reactions for metabolites set for keeping
                    keep_reac = [_.id for _ in model.metabolites.get_by_id(keep_meta.id).reactions]
                    # iterate over metabolites set for deletion
                    for del_meta_id in mnx[1]['id'].tolist()[1:]:
                        # retrieve reaction for metabolites set for deletion
                        del_reac = [_.id for _ in model.metabolites.get_by_id(del_meta_id).reactions]
                        # get intersection of reactions (keep + del)
                        reac_intersec = list(set(keep_reac) & set(del_reac))

                        # if intersection empty, metabolite is with a high probability indeed a duplicate
                        # Special case: NH3 / NH4
                        #    intersection does not have to be emtpy, know 'problem' caused by CarveMe
                        if len(reac_intersec) == 0 or all([re.search(NH_PATTERN,_) for _ in [keep_meta.id,del_meta_id]]):
                            # automated deletion is only advisable, if consistency can be retained
                            perform_deletion = True
                            with model as model_del:

                                # if the special case is detected ...
                                if all([re.search(NH_PATTERN,_) for _ in [keep_meta.id,del_meta_id]]):
                                    print(F'\tSpecial case -Duplicate NH4/NH3- detected.\n\tTrying to solve by additionally removing reactions containing both metabolites.')
                                    # ... remove reactions with nh3 and nh4 both present
                                    for del_reac_id in reac_intersec:
                                        # if objective_function is part of the set
                                        # automated deletion is (currently) not possible
                                        if del_reac_id == objective_function:
                                            perform_deletion = False
                                            break
                                        model_del.reactions.get_by_id(del_reac_id).remove_from_model()
                                    # set the metabolites to be deleted to be the one NOT in the objective functions
                                    # to avoid inconsistencies
                                    if del_meta_id in [_.id for _ in model_del.reactions.get_by_id(objective_function).metabolites]:
                                        temp = del_meta_id
                                        del_meta_id = keep_meta.id
                                        keep_meta = model_del.metabolites.get_by_id(temp)

                                # try replacing metabolite with the kept duplicate ...
                                for reac in model_del.metabolites.get_by_id(del_meta_id).reactions:

                                    reac.reaction = reac.reaction.replace(del_meta_id,keep_meta.id)

                                    # skip if objective function is found
                                    if reac.id == objective_function:
                                        continue

                                    # check if consistency is still intact
                                    balance_test = reac.check_mass_balance()
                                    if not reac.boundary and len(balance_test) > 0:
                                        # try fixing H-balance
                                        if 'H' in balance_test.keys():
                                            # ..............................
                                            # TODO:
                                            #    get H according to compartment
                                            #    current implementation relies heavily
                                            #    on 'correct' use input: compartment should have format C_c or c (C_p, p, C_e, e etc.)
                                            # ..............................
                                            reac_comp = reac.compartments.pop()[-1]
                                            if reac_comp == 'c':
                                                reac.subtract_metabolites({'h_c':balance_test['H']})
                                            elif reac_comp == 'p':
                                                reac.subtract_metabolites({'h_p':balance_test['H']})
                                            elif reac_comp == 'e':
                                                reac.subtract_metabolites({'h_e':balance_test['H']})
                                            else:
                                                perform_deletion = False
                                                break
                                        # ..............................
                                        # TODO:
                                        #    fix other possible problems
                                        # ..............................

                                        # finally, check balance again (continue only if fixed, else break)
                                        if len(reac.check_mass_balance()) > 0:
                                            perform_deletion = False
                                            break

                                    else:
                                        continue

                                # if not problems are found, duplicate is removed
                                if perform_deletion:
                                    model = model_del.copy()
                                    print(F'\tDuplicate metabolite {del_meta_id} found. Replaced with {keep_meta.id}.')
                                # if problems are not solvable, duplicate is kept and only reported
                                else:
                                    print(F'\tDuplicate metabolite {del_meta_id} found (duplicate to {keep_meta.id} based on annotation).\n\t\tAutomated deletion not possible due to problems with consistency.')

                        # else, metabolite is kept
                        #       since it might be an isomer, elongation, or other explanation
                        #       for the same annotation
                        else:
                            print(F'\tDuplicate metabolite {del_meta_id} found (duplicate to {keep_meta.id} based on annotation).\n\t\tKept, as reaction containing both metabolites was found.')


                # ... or only report duplicates
                else:
                    print(F'\tDuplicate metabolite(s) {", ".join(mnx[1]["id"].tolist())} found.')


    return model


def resolve_duplicates(model:cobra.Model, check_reac:bool=True, 
                       check_meta:Literal['default','exhaustive','skip']='default', 
                       replace_dupl_meta:bool=True, remove_unused_meta:bool=False, 
                       remove_dupl_reac:bool=True) -> cobra.Model:
    """Resolve and remove (optional) duplicate metabolites and reactions in the model.

    Args:
        - model (cobra.Model): 
            The model loaded with COBRApy.
        - check_reac (bool, optional): 
            Whether to check reactions for duplicates. 
            Defaults to True.
        - check_meta (Literal['default','exhaustive','skip'], optional): 
            Whether to check for duplicate metabolites. 
            Defaults to 'default'.
        - replace_dupl_meta (bool, optional): 
            Option to replace/remove duplicate metabolites. 
            Defaults to True.
        - remove_unused_meta (bool, optional): 
            Option to remove unused metabolites. 
            Defaults to False.
        - remove_dupl_reac (bool, optional): 
            Option to combine/remove duplicate reactions. 
            Defaults to True.

    Returns:
        cobra.Model: 
            The (edited) model.
    """

    # resolve duplicate metabolites
    if check_meta == 'default':
        # resolve duplicates starting with the metanetx.chemical database identifiers
        model = resolve_duplicate_metabolites(model, replace=replace_dupl_meta)
    elif check_meta == 'exhaustive':
        # resolve duplicates by starting at every database identifer one after another
        # note: bigg and sbo are skipped as sbo gives not much information and bigg is
        #       usually the one that differs (naming issue)
        anno_types = set()
        # get all database annotation types present in the model
        for m in model.metabolites:
            anno_types = anno_types | set(m.annotation.keys())
        for colname in [_ for _ in anno_types if not _ in ['bigg.metabolite','sbo']]:
            model = resolve_duplicate_metabolites(model,colname,replace=replace_dupl_meta)
    elif check_meta == 'skip':
        print('\tSkip check for duplicate metabolites.')
    else:
         warnings.warn(F'Warning: Unknown option for metabolites duplicate checking {check_meta}. Search for metabolite duplicates skipped.')

    # remove now unused metabolites
    if remove_unused_meta:
        model,removed = cobra.manipulation.delete.prune_unused_metabolites(model)
        print(F'\tThe following metabolites () have been removed: {", ".join([x.id for x in removed])}')

    # resolve duplicate reactions
    if check_reac:
        model = resolve_duplicate_reactions(model, based_on='reaction', remove_reac =remove_dupl_reac)

    return model





"""Collection of functions for setting up files and work environments.
"""

__author__ = 'Carolin Brune'

################################################################################
# requirements
################################################################################

import requests

from importlib.resources import files
from pathlib import Path
from tqdm import tqdm
from typing import Literal

################################################################################
# variables
################################################################################

PATH_MEDIA_CONFIG = files('refinegems.data.config').joinpath('media_config.yml') #: :meta hide-value: 

################################################################################
# functions
################################################################################

# --------------------------
# download databases / files
# --------------------------
# @TODO
def download_url(download_type:Literal['SwissProt gapfill'],
                 directory:str=None,k:int=10):
    """Download files necessary for certain functionalities of the toolbox from 
    the internet. 
    
    Currently available:
    
    - 'SwissProt gapfill': download files needed for the :py:class:`~refinegems.classes.gapfill.GeneGapFiller`
    

    Args:
        - dowload_type (Literal['SwissProt gapfill']): 
            Type of files to download.
        - directory (str, optional): 
            Path to a directory to save the downloaded files to. 
            Defaults to None.
        - k (int, optional): 
            Chunksize in kB. 
            Defaults to 10.

    Raises:
        - ValueError: Unknown database or file
    """
    
    # match URLS to type of database, that the user wants to download
    match download_type:
        case 'SwissProt gapfill':
            swissprot_api = 'https://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/complete/uniprot_sprot.fasta.gz'
            swissprot_mapping_api = 'https://rest.uniprot.org/uniprotkb/stream?fields=accession%2Cxref_brenda%2Cec&format=tsv&query=%28*%29+AND+%28reviewed%3Atrue%29'
            urls = {'SwissProt.fasta':swissprot_api, 'SwissProt_mapping.tsv':swissprot_mapping_api}
            # @TODO warning / prompt about size 
            #   1.: TSV with UniprotID, BRENDA and EC -7.7MB (26.07.2024)
            #   2.: FASTA with sequences ~280MB (26.07.2024)
            # @DISCUSSION contruct DIAMOND database in this step as well?
        case _:
            mes = f'Unknown database or file: {name}'
            raise ValueError(mes)

    # download each file
    for name, url in urls.items():
        r = requests.get(url, stream=True)
        filename = Path(directory, name) if directory else Path(name)
        
        # Check if Content-Length is available
        total_length = r.headers.get('Content-Length') # Make the progress bar
        if total_length is None:
            # Content-Length is missing, so we download without a progress bar
            # @TODO logging
            with open(filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=k * 1024):
                    if chunk:
                        f.write(chunk)
        else:
            total_length = int(total_length)
            with open(filename, 'wb') as f:
                pbar = tqdm(desc=f'Downloading {name}', 
                            unit="B", unit_scale=True, unit_divisor=1024, 
                            total=total_length)
                pbar.clear()
                for chunk in r.iter_content(chunk_size=k * 1024):
                    if chunk: # Filter out keep-alive new chunks
                        pbar.update(len(chunk)) # Update progress bar
                        f.write(chunk)
                pbar.close()

# ---------------------
# handling config files
# ---------------------


def download_config(filename:str='./my_config.yaml', type=Literal['media']):
    """Load a configuration file from the package and save a copy of it for the user to edit.

    Args:
        - filename (str, optional): 
            Filename to write the config to/save it under as. 
            Defaults to './my_config.yaml'.
        - type (Literal['media'], optional): 
            Type of configuration file to load.
            Can be 'media' for the media config file.
            Defaults to Literal['media'].
    """

    def copy_config_yaml(infile:str, outfile:str):
        """Helper function for :py:func:`download_config`. 
        Performs the actual download of a yaml file into a copy for the user to edit.

        Args:
            - infile (str): 
                Path to the file to copy.
            - outfile (str): 
                Path to write the copy to.
        """
        with open(infile, "r") as cfg_file, open(outfile, 'w') as cfg_out:
                for line in cfg_file:
                    cfg_out.write(line)

    # copy an examplary version of the config file for the user to edit it
    match type:

        # copy media config
        case 'media':           
            copy_config_yaml(PATH_MEDIA_CONFIG, filename)

        # type not found
        case _:
            raise ValueError(F'Unknown type of config file detected: {type}')


#!/usr/bin/env python
"""Most functions within this module were copied from the MEMOTE GitHub page and modified by Gwendolyn O. Döbel.

This module provides functions to normalise the biomass objective function(s).
"""

__author__ = "MEMOTE and Gwendolyn O. Döbel"

############################################################################
# requirements
############################################################################

import logging

from cobra import Model as cobraModel
from cobra import Reaction
from six import iteritems
from typing import Union

from ..utility.util import test_biomass_consistency, test_biomass_presence

############################################################################
# variables
############################################################################

############################################################################
# functions
############################################################################


# @DELETE - one the function below works?
def normalise_biomass(biomass: Reaction, current_sum: float) -> Reaction:
    """Normalises the coefficients according to current biomass weight to one g[CDW]

    Args:
        - biomass (Reaction): 
            Biomass function/reaction
        - current_sum (float): 
            Biomass weight calculated with sum_biomass_weight in g/mmol

    Returns:
        Reaction: 
            Biomass function/reaction with updated coefficients
    """
    metabs = biomass.metabolites # Get all metabolites
    
    # Normalise & update coefficients
    for (met, coef) in iteritems(metabs):
        metabs[met] = 1/current_sum * coef
    biomass._metabolites = metabs
    
    return biomass

def check_normalise_biomass(model: cobraModel, cycles:int=10) -> Union[cobraModel, None]:
    """
       1. Checks if at least one biomass reaction is present
       2. For each found biomass reaction checks if it sums up to 1g[CDW]
       3. Normalises the coefficients of each biomass reaction where the sum is not 1g[CDW] until the sum is 1g[CDW]
       4. Returns model with adjusted biomass function(s)
       

    Args:
        - model (cobraModel): 
            Model loaded with COBRApy
        - cycles (int, optional): 
            Maximal number of optiomisation cycles that will be run.
            Used to avoid endless optiomisation cycles.

    Returns:
        cobraModel: 
            COBRApy model with adjusted biomass functions
    """
    
    biomass_rxn = test_biomass_presence(model)
    
    if biomass_rxn:
        for bm_rxn in biomass_rxn:
            bm_weight = test_biomass_consistency(model, bm_rxn)
            if type(bm_weight) == str: logging.error(f'Reaction {bm_rxn}: {bm_weight}')
            else:
                c = 0 # counter to ensure it does not run endlessly
                while not ((1 - 1e-03) < bm_weight < (1 + 1e-06)) and c <= cycles:
                    #normalise_biomass(model.reactions.get_by_id(bm_rxn), bm_weight)
                    model.reactions.get_by_id(bm_rxn).__imul__(1/bm_weight) # -> Maybe add like this?
                    bm_weight = test_biomass_consistency(model, bm_rxn)
                    logging.info(f'For reaction \'{bm_rxn}\' the coefficients changed.')
                    c += 1
            
        return model
    
    else:
        logging.error(f'No biomass objective function was found in the provided model {model.id}.')
        return biomass_rxn



#!/usr/bin/env python
"""Variables, functions and more for the developement, extension and maintainance of 
the in-build database.

.. note:: 

   Some functionalities for handling and dealing with the media are in the 
   :py:mod:`~refinegems.classes.medium` module.
   
.. hint:: 

   Further functions for accessing the database can be found in the :py:mod:`~refinegems.utility.io`
   module, e.g. :py:func:`~refinegems.utility.io.load_a_table_from_database`
"""

__author__ = 'Gwendolyn O. Döbel und Carolin Brune'

################################################################################
# requirements
################################################################################

import io
import re
import sqlite3
import requests
import logging
import pandas as pd

from enum import Enum
from sqlite3 import Error
from tqdm import tqdm
from pathlib import Path
from importlib.resources import files
from typing import Union

################################################################################
# variables
################################################################################

PATH_TO_DB_FOLDER = files('refinegems.data.database') #: :meta hide-value: 
PATH_TO_DB = PATH_TO_DB_FOLDER.joinpath('data.db') #: :meta hide-value: 
VERSION_FILE = PATH_TO_DB_FOLDER.joinpath('current_bigg_db_version.txt') #: :meta hide-value: 
VERSION_URL = 'http://bigg.ucsd.edu/api/v2/database_version' #: :meta: 

mnx_db_namespace = {'reac_prop': ('https://www.metanetx.org/cgi-bin/mnxget/mnxref/reac_prop.tsv', 
                                  ['id','mnx_equation','reference','ec-code','is_balanced','is_transport']),
                    'reac_xref': ('https://www.metanetx.org/cgi-bin/mnxget/mnxref/reac_xref.tsv',
                                  ['source','id','description']),
                    'chem_prop': ('https://www.metanetx.org/cgi-bin/mnxget/mnxref/chem_prop.tsv',
                                  ['id','name','reference','formula','charge','mass','InChI','InChIKey','SMILES']),
                    'chem_xref': ('https://www.metanetx.org/cgi-bin/mnxget/mnxref/chem_xref.tsv',
                                  ['source','id','description'])
                    }

################################################################################
# functions
################################################################################

class ValidationCodes(Enum):
   """Validation codes for the database

      Args:
         - Enum (Enum): 
            Provided as input to get a number mapping for the codes
   """
   COMPLETE = 0,  # All tables are in data.db
   EMPTY = 1,  # data.db is either empty or incorrect
   BIGG = 2,  # Only BiGG tables are in data.db
   SBO_MEDIA = 3,  # Only SBO & Media tables are in data.db
   BIGG_SBO_MEDIA = 4,  # Only BiGG, SBO and media tables are in data.db 
   MODELSEED_COMPOUNDS = 5,  # Only ModelSEED compounds table is in data.db
   BIGG_MSEED_COMPOUNDS = 6,  # Only Bigg and ModelSEED compounds tables are in data.db
   SBO_MEDIA_MSEED_COMPOUNDS = 7  # Only SBO, media and ModelSEED compounds tables are in data.db
   
validation_messages = {
   ValidationCodes.COMPLETE: 
      'All tables in data up-to-date. Initialisation complete.',
   ValidationCodes.EMPTY: 
      'No table in data. An error must have occurred during initialisation.',
   ValidationCodes.BIGG: 
      'Data only contains the BiGG tables. Please check the remaining tables.',
   ValidationCodes.SBO_MEDIA: 
      'Data only contains the SBO and media tables. Please check the BiGG and ModelSEED tables.',
   ValidationCodes.BIGG_SBO_MEDIA: 
      'Data only contains the BiGG, SBO and media tables. Please check the ModelSEED table.',
   ValidationCodes.MODELSEED_COMPOUNDS: 
      'Data only contains the ModelSEED table. Please check the BiGG, SBO and media tables.',
   ValidationCodes.BIGG_MSEED_COMPOUNDS: 
      'Data only contains the BiGG and ModelSEED tables. Please check the SBO and media tables.',
   ValidationCodes.SBO_MEDIA_MSEED_COMPOUNDS: 
      'Data only contains the SBO, media and ModelSEED tables. Please check the BiGG tables.'
}


def is_valid_database(db_cursor: sqlite3.Cursor) -> int:
   """Verifies if database has:

      - 2 tables with names 'bigg_metabolites' & 'bigg_reactions'
      - 2 tables with names 'bigg_to_sbo' & 'ec_to_sbo'
      - 6 tables with names 'medium', 'substance', 'substance2db' & 'medium2substance', 'subset' & 'subset2substance'
      - 1 table with name 'modelseed_compounds'

   Args:
      - db_cursor (sqlite3.Cursor): 
         Cursor from open connection to the database (data.db)

   Returns:
      int: 
         Corresponding to one of the ValidationCodes
   """
   print('Verifying database...')
   
   # Fetches the table names as string tuples from the connected database
   db_cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
   tables = [string[0] for string in db_cursor.fetchall()]
   
   bigg_tables_contained = len([s for s in tables if re.match('^bigg_(?!to)(.*?)', s, re.IGNORECASE)]) == 2
   sbo_tables_contained = len([s for s in tables if re.match('(.*?)_sbo$', s, re.IGNORECASE)]) == 2
   media_tables_contained = len([s for s in tables if re.match('^medium(.*?)|^substance(.*?)|^subset(.*?)', s, re.IGNORECASE)]) == 6
   sbo_media_tables_contained = sbo_tables_contained and media_tables_contained  # These can only occur together
   modelseed_cmpd_tbl_contained = len([s for s in tables if s == 'modelseed_compounds']) == 1
   
   bigg_sbo_media_tbls_contained = bigg_tables_contained and sbo_media_tables_contained
   bigg_modelseed_cmpd_tbls_contained = bigg_tables_contained and modelseed_cmpd_tbl_contained
   sbo_media_modelseed_cmpd_tbls_contained = sbo_media_tables_contained and modelseed_cmpd_tbl_contained
   all_tables_contained = bigg_sbo_media_tbls_contained and modelseed_cmpd_tbl_contained
   
   if all_tables_contained: return ValidationCodes.COMPLETE
   elif bigg_modelseed_cmpd_tbls_contained: return ValidationCodes.BIGG_MSEED_COMPOUNDS
   elif sbo_media_modelseed_cmpd_tbls_contained: return ValidationCodes.SBO_MEDIA_MSEED_COMPOUNDS
   elif bigg_sbo_media_tbls_contained: return ValidationCodes.BIGG_SBO_MEDIA
   elif bigg_tables_contained: return ValidationCodes.BIGG
   elif sbo_media_tables_contained: return ValidationCodes.SBO_MEDIA
   elif modelseed_cmpd_tbl_contained: return ValidationCodes.MODELSEED_COMPOUNDS
   else: return ValidationCodes.EMPTY


def create_sbo_media_database(db_cursor: sqlite3.Cursor):
   """Creates the SBOannotator database with 2 tables ('bigg_to_sbo' & 'ec_to_sbo') from file './data/database/sbo_mapping_db.sql'
   and the media database with 4 tables ('medium', 'substance', 'substance2db', 'medium2substance') from file './data/database/media_db.sql'

   Args:
      - db_cursor (sqlite3.Cursor): 
         Cursor from open connection to the database (data.db)
   """
   print('Adding SBO tables...')
   with open(files('refinegems.data.database').joinpath('sbo_mapping_db.sql')) as schema:
      db_cursor.executescript(schema.read())
   
   
   db_cursor.executescript('''
                           DROP TABLE IF EXISTS susbtance;
                           DROP TABLE IF EXISTS substance2db;
                           DROP TABLE IF EXISTS medium;
                           DROP TABLE IF EXISTS medium2substance;
                           ''')
   
   print('Adding media tables...')
   with open(files('refinegems.data.database').joinpath('media_db.sql')) as schema:
      db_cursor.executescript(schema.read())


def update_bigg_db(latest_version: str, db_connection: sqlite3.Connection) -> dict:
    """Updates the BiGG tables 'bigg_metabolites' & 'bigg_reactions' within a database (data.db)

    Args:
      - latest_version (str): 
         String containing the Path to a file with the latest version of the BiGG database
      - db_connection (sqlite3.Connection): 
         Open connection to the database (data.db)
    """

    def get_database_links_info_per_row(row:pd.Series):
        """For a single row of the dataframe, extract the database identifier from the collection of database links.

        Args:
            - ow (pd.Series): 
               The database links for a single row of the dataframe.

        Returns:
            dict: 
               A dictionary with the databses as keys and the IDs as values.
        """

        database_ids = {}
        if isinstance(row, str):
            links = row.split(';')
            for link in links:
                key, value = link.split(':',1)
                key = key.strip()
                value = value.rsplit('/',1)[1].strip()
                value = re.sub(r'^(?i)meta:','',value)
                if key in database_ids.keys():
                    database_ids[key].append(value)
                else:
                    database_ids[key] = [value]

        for k in database_ids.keys():
            database_ids[k] = ', '.join(database_ids[k])

        return database_ids

    def get_database_links(data:pd.DataFrame) -> pd.DataFrame:
        """For the dataframe, extract the database IDs from the database_links column
        into separate column containing the IDs and not the links.

        Args:
            - data (pd.DataFrame): 
               The input dataframe.

        Returns:
            pd.DataFrame: 
               The edited dataframe
        """
    
        data = data.join(pd.DataFrame([get_database_links_info_per_row(row) for row in data['database_links']]))
        data.drop(columns=['database_links', 'model_list'], axis=1, inplace=True)

        return data

    print('Adding BiGG tables...')
    
    # Store currently used version
    with open(VERSION_FILE, 'w') as file:
        file.write(latest_version)
    
    # Create BiGG metabolites table
    BIGG_MODELS_METABS_URL = 'http://bigg.ucsd.edu/static/namespace/bigg_models_metabolites.txt'
    bigg_models_metabs = requests.get(BIGG_MODELS_METABS_URL).text
    bigg_models_metabs_df = pd.read_csv(io.StringIO(bigg_models_metabs), dtype=str, sep='\t')
    bigg_models_metabs_df.rename(columns={'bigg_id': 'id'}, inplace=True)
    
    bigg_models_metabs_df = get_database_links(bigg_models_metabs_df)
    
    bigg_id_duplicates = bigg_models_metabs_df.duplicated(subset=['id'], keep=False)
    bigg_id_duplicates_df = bigg_models_metabs_df[bigg_id_duplicates]
    bigg_id_duplicates_set = set(bigg_id_duplicates_df['id'].tolist())
   
    bigg_models_metabs_df[~bigg_id_duplicates].to_sql(
        'bigg_metabolites', db_connection, 
        if_exists='replace', index=False, 
        dtype={'id':'TEXT PRIMARY KEY'}
        )
    
    if bigg_id_duplicates_set:
        logging.warning(
            'The BiGG metabolite table contains the following '
            f'{len(bigg_id_duplicates_set)} duplicate(s):\n'
            f'{bigg_id_duplicates_set}\n'
            'Duplicate(s) are completely removed from the table.'
            )

    # Create BiGG reactions table
    BIGG_MODELS_REACS_URL = 'http://bigg.ucsd.edu/static/namespace/bigg_models_reactions.txt'
    bigg_models_reacs = requests.get(BIGG_MODELS_REACS_URL).text
    bigg_models_reacs_df = pd.read_csv(io.StringIO(bigg_models_reacs), dtype=str, sep='\t')
    bigg_models_reacs_df.rename(columns={'bigg_id': 'id'}, inplace=True)

    bigg_models_reacs_df = get_database_links(bigg_models_reacs_df)

    bigg_models_reacs_df.to_sql(
        'bigg_reactions', db_connection, 
        if_exists='replace', index=False, 
        dtype={'id':'TEXT PRIMARY KEY'}
        )
   

def get_latest_bigg_databases(db_connection: sqlite3.Connection, is_missing: bool=True):
   """Gets the latest BiGG tables for metabolites & reactions if:

      - No version file is locally available
      - The version in the local version file is NOT the latest
      - No BiGG tables currently exist in the database

   Args:
      - db_connection (sqlite3.Connection): 
         Open connection to the database (data.db)
      - is_missing (bool, optional): 
         True if no BiGG tables are in the database. Defaults to True.
   """
   # Check if BiGG database had an update
   LATEST_VERSION = requests.get(VERSION_URL).json()['bigg_models_version']
   
   if not Path.exists(VERSION_FILE) or is_missing:
      update_bigg_db(LATEST_VERSION, db_connection)
      
   else:
      with open(VERSION_FILE, 'r') as file:
         version = file.readline().strip()
         
      if version != LATEST_VERSION:
         update_bigg_db(LATEST_VERSION, db_connection)
 
 
def get_modelseed_compounds_database(db_connection: sqlite3.Connection):
   """Retrieves the compounds table from ModelSEED from the respective GitHub repository

   Args:
      - db_connection (sqlite3.Connection): 
         Open connection to the database (data.db)
   """
   print('Adding the ModelSEED compounds table...')
   MODELSEED_COMPOUNDS_URL = 'https://raw.githubusercontent.com/ModelSEED/ModelSEEDDatabase/master/Biochemistry/compounds.tsv'
   modelseed_compounds = requests.get(MODELSEED_COMPOUNDS_URL).text
   modelseed_df = pd.read_csv(io.StringIO(modelseed_compounds), sep='\t', dtype={'linked_compound': str}) #'mass': float, 
   modelseed_df.to_sql(
      'modelseed_compounds', db_connection, 
      if_exists='replace', index=False, 
      dtype={
         'id':'TEXT PRIMARY KEY',
         'mass': 'NUMERIC',
         'linked_compound': 'TEXT'
         }
      )
    
         
def initialise_database():
   """Initialises/updates the database (data.db)

      After initialisation the database contains:

      - 2 tables with names 'bigg_metabolites' & 'bigg_reactions'
      - 2 tables with names 'bigg_to_sbo' & 'ec_to_sbo'
      - 6 tables with names 'medium', 'substance', 'medium2substance', 'substance2db', 'subset' & 'subset2substance'
      - 1 table with name 'modelseed_compounds' 
   """
   # Initialise empty connection
   con = None

   print('Initialising database...')
   
   # Try to open connection & get cursor
   try:
      con = sqlite3.connect(PATH_TO_DB)
      cursor = con.cursor()
      
      validity_code = is_valid_database(cursor)
      
      if validity_code == ValidationCodes.BIGG:
         print('Only BiGG tables contained in database.')
         create_sbo_media_database(cursor)
         get_modelseed_compounds_database(con)
      
      elif validity_code == ValidationCodes.SBO_MEDIA:
         print('Only SBO and media tables contained in database.')
         get_latest_bigg_databases(con)
         get_modelseed_compounds_database(con)
         
      elif validity_code == ValidationCodes.MODELSEED_COMPOUNDS:
         print('Only ModelSEED compounds table contained in database.')
         create_sbo_media_database(cursor)
         get_latest_bigg_databases(con)
         
      elif validity_code == ValidationCodes.BIGG_SBO_MEDIA:
         print('Only BiGG, SBO and media tables contained in database.')
         get_modelseed_compounds_database(con)
         
      elif validity_code == ValidationCodes.BIGG_MSEED_COMPOUNDS:
         print('Only BiGG and ModelSEED compounds tables contained in database.')
         create_sbo_media_database(cursor)
         
      elif validity_code == ValidationCodes.SBO_MEDIA_MSEED_COMPOUNDS:
         print('Only SBO, media and ModelSEED compounds tables contained in database.')
         get_latest_bigg_databases(con)
         
      elif validity_code == ValidationCodes.EMPTY:
         print('Incorrect or empty database. Initialise database with required tables...')
         create_sbo_media_database(cursor)
         get_latest_bigg_databases(con)
         get_modelseed_compounds_database(con)
         
      elif validity_code == ValidationCodes.COMPLETE:
         print('Verifying if BiGG tables are up-to-date...')
         get_latest_bigg_databases(con, False)
      
   except Error as e:
      print(e)
   finally:
      if con: con.close()
      
      # Validate initialised database
      con = sqlite3.connect(PATH_TO_DB)
      cursor = con.cursor()
      validity_code = is_valid_database(cursor)
      print(validation_messages.get(validity_code))
      con.close()


# about MetaNetX Information
# --------------------------
# DISCLAIMER:
# Database information from MetaNetX
# distributed under https://creativecommons.org/licenses/by/4.0/
# Citation: MetaNetX/MNXref: unified namespace for metabolites and biochemical reactions in the context of metabolic models
#           Sébastien Moretti, Van Du T Tran, Florence Mehl, Mark Ibberson, Marco Pagni
#           Nucleic Acids Research (2021), 49(D1):D570-D574

# @TODO time warning + progress bar (waiting w/o info is tedious)
def update_mnx_namespaces(db:Union[Path,str]=PATH_TO_DB, chunksize:int=1):
   """Add or update the MetaNetX namespace to/in a database. 
   
   Args:
      - db (Union[Path,str],optional):
         Path to a database to add the namespace to.
         Defaults to the in-build database.
      - chunksize (int, optional):
         Size of the chunk (in kB) to download at once.
         Defaults to 1.
   """

   con = sqlite3.connect(db)
   for name,values in mnx_db_namespace.items():
      link,colnames = values
      mnx_table = []
      for chunk in tqdm(pd.read_csv(link, sep='\t', comment='#', 
                                    names=colnames, 
                                    chunksize=chunksize*1024), 
                                    desc=f'Downloading {name}',
                                    unit='B'):# progress bar will not work -> no totel length info
         mnx_table.append(chunk) 

      match name:
         # Reaction property table
         case 'reac_prop':
               total_len = sum([len(_) for _ in mnx_table])
               with tqdm(total=total_len, unit='entries', 
                        desc='Add to DB') as pbar:
                  for i,chunk in enumerate(mnx_table):
                     if i == 0:
                           exists = 'replace'
                     else:
                           exists = 'append'
                     chunk.to_sql(
                           'mnx_'+name, con, 
                           if_exists=exists, index=False, 
                           dtype={'id':'TEXT PRIMARY KEY'}
                           )
                     pbar.update(len(chunk)) 
      
         # Reaction cross-reference table
         case 'reac_xref':
               cursor = con.cursor()
               cursor.execute('DROP TABLE IF EXISTS mnx_reac_xref')
               empty_table = """ CREATE TABLE mnx_reac_xref (
                                 source TEXT,
                                 id TEXT,
                                 description TEXT,
                                 CONSTRAINT PK_mnx_reac_xref PRIMARY KEY (source,id)
                                 FOREIGN KEY(id) REFERENCES mnx_reac_prop(id)
                           );
                           """
               cursor.execute(empty_table)
               total_len = sum([len(_) for _ in mnx_table])
               with tqdm(total=total_len, unit='entries',
                        desc='Add to DB') as pbar:
                  for i,chunk in enumerate(mnx_table):
                     chunk.to_sql(
                           'mnx_'+name, con, 
                           if_exists='append', index=False
                           )
                     pbar.update(len(chunk))
               
         # Metabolite properties table
         case 'chem_prop':
               total_len = sum([len(_) for _ in mnx_table])
               with tqdm(total=total_len, unit='entries',
                        desc='Add to DB') as pbar:
                  for i,chunk in enumerate(mnx_table):
                     if i == 0:
                           exists = 'replace'
                     else:
                           exists = 'append'
                     chunk.to_sql(
                           'mnx_'+name, con, 
                           if_exists=exists, index=False, 
                           dtype={'id':'TEXT PRIMARY KEY'}
                           )
                     pbar.update(len(chunk))
                     
         # Metabolite reference table
         case 'chem_xref':
               total_len = sum([len(_) for _ in mnx_table])
               cursor = con.cursor()
               cursor.execute('DROP TABLE IF EXISTS mnx_chem_xref')
               empty_table = """ CREATE TABLE mnx_chem_xref (
                                 source TEXT,
                                 id TEXT,
                                 description TEXT,
                                 CONSTRAINT PK_mnx_chem_xref PRIMARY KEY (source,id)
                                 FOREIGN KEY(id) REFERENCES mnx_chem_prop(id)
                           );
                           """
               cursor.execute(empty_table)
               with tqdm(total=total_len, unit='entries',
                        desc='Add to DB') as pbar:
                  for i,chunk in enumerate(mnx_table):
                     chunk.to_sql(
                           'mnx_'+name, con, 
                           if_exists='append', index=False
                           )
                     pbar.update(len(chunk))
   con.close()
   

# @TEST
def reset_database(database:Path|str=PATH_TO_DB):
    """Remove tables for certain databases to allow pushing of the database
    to GitHub (reduce size).

    Args:
        - database (Path | str, optional): 
            Path to the database. Defaults to PATH_TO_DB, the in-build database.
    """
    # establish connection
    con = sqlite3.connect(database)
    cursor = con.cursor()
    # remove MetaNetX tables
    cursor.execute('DROP TABLE IF EXISTS mnx_chem_xref')
    cursor.execute('DROP TABLE IF EXISTS mnx_reac_xref')
    cursor.execute('DROP TABLE IF EXISTS mnx_chem_prop')
    cursor.execute('DROP TABLE IF EXISTS mnx_reac_prop')
    # close connection
    con.close()



"""Identify, report and solve energy generating cycles (EGCs).
"""

__author__ = 'Carolin Brune and Tobias Fehrenbach'

################################################################################
# requirements
################################################################################

from ..analysis.growth import MIN_GROWTH_THRESHOLD, set_bounds_to_default
from .medium import Medium
from ..utility.util import test_biomass_presence

import cobra
import pandas as pd
import warnings

from functools import partial
from multiprocess import Pool
from tqdm import tqdm
from typing import Literal

################################################################################
# variables
################################################################################

DISSIPATION_RXNS = {
    "ATP": {"ATP [Adenosine triphosphate]": -1, "Water [H2O]": -1, "ADP [Adenosine diphosphate]": 1, "Hydrogen [H(+)]": 1, "Phosphate [PO4(3-)]": 1},
    "CTP": {"CTP [Cytidine triphosphate]": -1, "Water [H2O]": -1, "CDP [Cytidine diphosphate]": 1, "Hydrogen [H(+)]": 1, "Phosphate [PO4(3-)]": 1},
    "GTP": {"GTP [Guanosine triphosphate]": -1, "Water [H2O]": -1, "GDP [Guanosine diphosphate]": 1, "Hydrogen [H(+)]": 1, "Phosphate [PO4(3-)]": 1},
    "UTP": {"UTP [Uridine triphosphate]": -1, "Water [H2O]": -1, "UDP [Uridine diphosphate]": 1, "Hydrogen [H(+)]": 1, "Phosphate [PO4(3-)]": 1},
    "ITP": {"ITP [Inosine triphosphate]": -1, "Water [H2O]": -1, "IDP [Inosine diphosphate]": 1, "Hydrogen [H(+)]": 1, "Phosphate [PO4(3-)]": 1},
    "NADH": {"NADH [reduced Nicotinamide adenine dinucleotide]": -1, "Hydrogen [H(+)]": 1, "NAD [oxidized Nicotinamide adenine dinucleotide]": 1},
    "NADPH": {"NADPH [reduced Nicotinamide adenine dinucleotide phosphate]": -1, "Hydrogen [H(+)]": 1, "NADP [oxidized Nicotinamide adenine dinucleotide phosphate]": 1},
    "FADH2": {"FADH2 [reduced Flavin adenine dinucleotide]": -1, "Hydrogen [H(+)]": 2, "FAD [oxidized Flavin adenine dinucleotide]": 1},
    "FMNH2": {"FMNH2 [reduced Flavin mononucleotide]": -1,  "Hydrogen [H(+)]": 2, "FMN [oxidized Flavin mononucleotide]": 1},
    "Q8H2": {"Ubiquinone-8": -1, "Hydrogen [H(+)]": 2, "Ubiquinol-8": 1},
    "MQL8": {"Menaquinone-8": -1, "Hydrogen [H(+)]": 2, "Menaquinol-8": 1},
    "DMMQL8": {"2-Demethylmenaquinone-8": -1, "Hydrogen [H(+)]": 2, "2-Demethylmenaquinol-8": 1},
    "ACCOA": {"Acetyl-CoA": -1, "Water [H2O]": -1, "Hydrogen [H(+)]": 1, "Acetate [Acetic acid]": 1, "Coenzyme A": 1},
    "GLU": {"D-Glucose": -1, "Water [H2O]": -1, "2-Oxoglutarate [Oxoglutaric acid]": 1, "Ammonia": 1, "Hydrogen [H(+)]": 2},
    "PROTON": {"Hydrogen [H(+)]": 1, "Hydrogen [H(+)] transported": -1}
} #: :meta: 

EGC_SCORING_MATRIX = {'MR':1, 'RB':3, 'RF':3,'RM':6} #: :meta: 

################################################################################
# classes
################################################################################

class EGCSolver():
    """Parent class for the EGC solvers with generally useful functions and 
    attributes. Can only be used to find, not solve EGCs directly.

    Attributes:
        - theshold: Float describing the cutoff, under which the model
            will no longer considered to be growing.
            Defaults to the MIN_GROWTH_THRESHOLD set in the growth module.
        - limit: Sets the maximal number of cores to be used.
            Defaults to 2.
        - chunksize: Chunksize to use for multiprocessing.
            Defaults to 12.
    """

    def __init__(self, threshold:float = MIN_GROWTH_THRESHOLD, 
                 limit:int=8, chunksize:int=12) -> None:
        
        # 'biological' attributes
        self.threshold = threshold

        # purely computational attributes
        self.limit = limit
        self.chunksize = chunksize

    # general functions
    # -----------------
    def check_metab_integration(self, metabolites: dict[str: int], model: cobra.Model,
                                 metab_info:Medium, namespace:Literal['BiGG']='BiGG',
                                 compartment:list=['c','p']) -> None|dict:
            """Check if the metabolites of a reactions are in the model.
            If yes, return the dictionary of metabolites (their IDs in the model) to the factors.
            If no, return None

            Args:
                - metabolites (dict[str: int]): 
                    Metabolites mapped to factors.
                - model (cobra.Model): 
                    The model loaded with COBRApy.
                - metab_info (Medium): 
                    Information about the metabolites from the database,
                    in for of a Medium object.
                - namespace (Literal['BiGG'], optional): 
                    String for the namespace used in the model. 
                    Current options include 'BiGG'.
                    Defaults to 'BiGG'.
                - compartment (list, optional): 
                    List of length 2 with the names of the 
                    compartments for the dissipations reactions. 
                    Defaults to ['c','p'].

            Returns:
                (1) Case: metabolites not found
                        None: 
                            nothing to return
                
                (2) Case: found  
                        dict: 
                            The mapping of IDs of the metabolites to the factors.
            """

            found_ids = {}

            c1_metab = [_.id for _ in model.metabolites if _.compartment == compartment[0]]
            c2_metab = [_.id for _ in model.metabolites if _.compartment == compartment[1]]

            for meta in list(metabolites.keys()):
                # get metabolite database annotations
                pos_ids = metab_info.substance_table[metab_info.substance_table['name']==meta][['db_type','db_id']]
                # check namespace availability 
                if not any(namespace in _ for _ in list(pos_ids['db_type'])):
                    return None
                # check metabolite availability in model
                id = pos_ids[pos_ids['db_type'].str.contains(namespace)]['db_id']
                for i in id:
                    match namespace:
                        # BiGG ID needs the compartment with a '_' as suffix
                        case 'BiGG':
                            if len(metabolites) == 2: # special case proton
                                i_p = i + '_' + compartment[1]
                                i_c = i + '_' + compartment[0]
                                if i_p not in c2_metab or i_c not in c1_metab:
                                    return None
                                else:
                                    found_ids[i_p] = metabolites['Hydrogen [H(+)] transported']
                                    found_ids[i_c] = metabolites['Hydrogen [H(+)]']
                                    return found_ids
                            else:
                                i += '_' + compartment[0]
                        # No namespacce given
                        case _:
                            mes = f'Unknown namespace or no namespace given: {namespace}'
                            warnings.warn(mes)
                            return None

                    if i not in c1_metab:
                        return None
                    else:
                        found_ids[i] = metabolites[meta]
                        break 

            return found_ids


    def add_DISSIPATIONRXNS(self, model: cobra.Model,
                        namespace:Literal['BiGG']='BiGG',
                        compartment:list=['c','p']) -> cobra.Model:
        """Add the dissipation reactions a model.

        Args:
            - model (cobra.Model): 
                A model loaded with COBRApy.
            - namespace (Literal['BiGG'], optional): Namespace of the model. 
                Defaults to 'BiGG'.
            - compartment (list, optional): 
                List of length 2 with the names of the compartments for the dissipations reactions. 
                Defaults to ['c','p'].

        Returns:
            cobra.Model: 
                The edited model.
        """
    
        # retrieve information about dissipation reaction metabolites
        metab_info = Medium('dissipation reaction metabolites')
        metab_info.add_subset('DiReM')
        
        # add dissipations reactions to model
        for name, metabolites in DISSIPATION_RXNS.items():
            if "DISSI_"+name not in model.reactions:
                ids_in_namespace = self.check_metab_integration(metabolites, model,
                                        metab_info, namespace=namespace,
                                        compartment=compartment)
                if ids_in_namespace:
                    rea = cobra.Reaction(name=name, id="DISSI_"+name)
                    model.add_reactions([rea])
                    rea.add_metabolites(ids_in_namespace)

        return model


    def limit_bounds(self, model: cobra.Model):
        """Limits upper and lower bounds of

        - exchange reactions to (0, 0)
        - reversible reactions to (-1, 1)
        - irreversible reactions to (0, 1)

        Excludes dissipation reactions.

        Args:
            - model (cobra.Model): 
                COBRApy model
        """
        external_comp = cobra.medium.find_external_compartment(model)
        # set fluxes for each reaction within model
        for rea in model.reactions:
            # except dissipation reactions
            if "DISSI_" in rea.id:  
                continue
            # turn off exchange reactions
            elif cobra.medium.is_boundary_type(rea,'exchange',external_comp):
                rea.bounds = (0.0, 0.0)
            # limit reversible reactions to [-1, 1] -> flux 0.1
            elif rea.reversibility: 
                rea.bounds = (-1.0, 1.0)
            # limit irreversible reactions to [0, 1] -> flux 0.1
            else: 
                rea.bounds = (0.0, 1.0)


    # find EGCs
    # ---------

    def find_egcs(self, model:cobra.Model, with_reacs:bool=False,
                  namespace:Literal['BiGG']='BiGG',
                  compartment:list=['c','p']) -> list|tuple:
        """Find the EGCs in a model - if exsistend.

        Args:
            - model (cobra.Model): 
                The model loaded with COBRApy.
            - with_reacs (bool, optional): 
                Option to either only return the names
                of the found EGC or additionally also the reactions, which 
                show fluxes during testing. 
                Defaults to False.
            - namespace (Literal['BiGG'], optional): 
                String for the namespace used in the model. 
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional): 
                List of length 2 with the names of the 
                compartments for the dissipations reactions. 
                Defaults to ['c','p'].

        Returns:
            (1) Case: ``with_reacs = False``
                    list: 
                        List of found EGC names.

            (2) Case: ``with_reacs = True``
                    tuple: 
                        tuple of (1) dictionary & (2) list:

                        (1) dict: dictionary of the EGCs 
                        (2) list: their reactions that showed fluxes and the objective values of the test.
        """
        # add 15 energy dissipation reactions
        with model as mod_model:  

            # ensure no model modifications
            mod_model = self.add_DISSIPATIONRXNS(mod_model,namespace,compartment)

            # set fluxes for each reaction within model
            self.limit_bounds(mod_model)

            # set each dissipation reaction as objective & optimize
            # options: just return the names of the found cycles
            if not with_reacs:
                egcs = []
                for name in DISSIPATION_RXNS.keys():
                    rea_id = "DISSI_"+name
                    if rea_id in mod_model.reactions:
                        mod_model.objective = "DISSI_"+name
                        solution = mod_model.optimize()

                        if solution.objective_value > MIN_GROWTH_THRESHOLD:  # optimization > 0 --> EGC detected
                            egcs.append(name)

                return egcs
            # option 2: report cycles + problematic reactions with fluxes
            else:
                egc_reactions = {}
                obj_vals = {}
                for name in DISSIPATION_RXNS.keys():
                    rea_id = "DISSI_"+name
                    if rea_id in mod_model.reactions:
                        mod_model.objective = "DISSI_"+name
                        solution = mod_model.optimize()
                        fluxes = solution.fluxes
                        objval = solution.objective_value

                        if objval > MIN_GROWTH_THRESHOLD:  # optimization > 0 --> EGC detected
                            obj_vals[name] = objval
                            egc_reactions[name] = {}

                            for rea, flux in fluxes.items():
                                # cutoff for flux 
                                if  (flux > MIN_GROWTH_THRESHOLD or flux < -1.0 * MIN_GROWTH_THRESHOLD) and not rea.startswith("DISSI"):
                                    egc_reactions[name][rea] = flux

                return (egc_reactions, obj_vals)

    # Helper functions for testing during solving process
    # ---------------------------------------------------

    def egcs_removed(self, model: cobra.Model, starting_egcs: dict,
                     namespace:Literal['BiGG']='BiGG', 
                     compartment:list=['c','p']) -> list:
        """Compare a list of previously found EGCs to the current 
        EGCs in the model.

        Args:
            - model (cobra.Model): 
                The model loaded with COBRApy after
                a try of solving the EGCs.
            - starting_egcs (dict): 
                List of EGCs before trying to solve them.
            - namespace (Literal['BiGG'], optional): 
                String for the namespace used in the model. 
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional): 
                List of length 2 with the names of the compartments for the dissipations reactions. 
                Defaults to ['c','p'].

        Returns:
            list: 
                List of newly removed EGCs.
        """

        current_egcs = self.find_egcs(model, namespace=namespace, compartment=compartment)

        removed = [_ for _ in starting_egcs.keys() if _ not in current_egcs]
        added = [_ for _ in current_egcs if _ not in starting_egcs.keys()]

        if len(added) > 0:
            
            return []
        else: 
            return removed
    



class GreedyEGCSolver(EGCSolver):
    """EGC solver that finds a good solution (greedy) based on modifications
    to single reactions.

    Workflow:

    - identify existing EGCs
    - test, if EGCs can be solved using single modifications of reactions

        - possible modifications:

            - deletion (RM)
            - set reversible (MR)
            - remove backward (forward only) (RB)
            - remove forward (backward only) (RF)

    - find a good - not optimal - combination of reactions, that solve
      the maximum number of EGCs that can be solved this way
    - apply solution to the model
    - report remaining EGCs, score and reactions used for solution

    Attributes:
        - all attributes of the base class :py:class:`refinegems.classes.egcs.EGCSolver`
        - scoring_matrix: 
            Dictionary of the changes (RM, MR, RF, RB) against
            Integers describing the penalty scores.
    """

    def __init__(self, scoring_matrix:dict = EGC_SCORING_MATRIX, **kwargs) -> None:
        super().__init__(**kwargs)
        self.scoring_matrix = scoring_matrix
        

    # find modifications to solve single EGCs
    # ---------------------------------------
    
    def check_egc_growth(self, reac: cobra.Reaction, model: cobra.Model, 
                bounds:tuple, starting_egcs:dict,
                namespace:Literal['BiGG']='BiGG',
                compartment:list=['c','p']) -> list|None:
        """Check EGC removal and growth of a model when chaning the bounds
        of a single reaction.

        Args:
            - reac (cobra.Reaction): 
                The reaction to change
            - model (cobra.Model): 
                The model (COBRApy) to manipulate.
            - bounds (tuple): 
                The new reactions bounds.
            - starting_egcs (dict): 
                Dict of the original EGCs found in the model.
            - namespace (Literal['BiGG'], optional): 
                String for the namespace used in the model. 
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional): List of length 2 with the names of the 
                compartments for the dissipations reactions. 
                Defaults to ['c','p'].

        Returns:
            (1) Case if EGCs removed  
                    list: 
                        List of EGCs that can be removed with the change 
            
            (2) Case no removal possible  
                    None: 
                        no return
        """

        # set new reaction bounds
        reac.bounds = bounds
        # check egcs removal and growth
        egc_test = self.egcs_removed(model, starting_egcs,namespace,compartment)
        res = model.optimize()
        if len(egc_test) > 0 and res.status == 'optimal' and res.objective_value > self.threshold:
            return egc_test
        else:
            None



    def test_modifications(self,reaction: cobra.Reaction, model: cobra.Model, 
                        present_egc: dict, namespace:Literal['BiGG']='BiGG',
                        compartment:list=['c','p']) -> dict:
        """Tries four cases for a Reaction

        1. if reaction is not reversible -> make reaction reversible (MR)
        2. limit backward reaction (RB)
        3. limit forward reaction (RF)
        4. "delete" reaction by setting fluxes to 0 (RM)
    
        -> for each case the EGCs which are present in the model are checked if they are removed
        
        -> if EGCs are removed we check if the model still grows on optimal medium
        
        => When both limitations are True reaction is saved to corresponding dictionary

        
        Args:
            - reaction (cobra.Reaction): 
                Reaction from a cobra.Model
            - model (cobra.Model): 
                The corresponding GEM loaded with cobrapy
            - present_egc (dict): 
                Dictionary of present EGCs {"egc": {}} -> EGCs are keys
            - namespace (Literal['BiGG'], optional): 
                String for the namespace used in the model. 
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional): 
                List of length 2 with the names of the 
                compartments for the dissipations reactions. 
                Defaults to ['c','p'].

        Returns:
            dict: 
                {"egc": {"MR":[potential_solutions],
                "RB":[potential_solutions],
                "RF":[potential_solutions],
                "RM":[potential_solutions]}}
        """
        results = {
                "ATP": {"MR": [], "RB": [], "RF": [], "RM": []},
                "CTP": {"MR": [], "RB": [], "RF": [], "RM": []},
                "GTP": {"MR": [], "RB": [], "RF": [], "RM": []},
                "UTP": {"MR": [], "RB": [], "RF": [], "RM": []},
                "ITP": {"MR": [], "RB": [], "RF": [], "RM": []},
                "NADH": {"MR": [], "RB": [], "RF": [], "RM": []},
                "NADPH": {"MR": [], "RB": [], "RF": [], "RM": []},
                "FADH2": {"MR": [], "RB": [], "RF": [], "RM": []},
                "FMNH2": {"MR": [], "RB": [], "RF": [], "RM": []},
                "Q8H2": {"MR": [], "RB": [], "RF": [], "RM": []},
                "MQL8": {"MR": [], "RB": [], "RF": [], "RM": []},
                "DMMQL8": {"MR": [], "RB": [], "RF": [], "RM": []},
                "ACCOA": {"MR": [], "RB": [], "RF": [], "RM": []},
                "GLU": {"MR": [], "RB": [], "RF": [], "RM": []},
                "PROTON": {"MR": [], "RB": [], "RF": [], "RM": []}
                }
    
        # skip exchange reactions
        if reaction.boundary:
            return None
        
        # test modifications of the reaction
        with model as m:
            original_bounds = reaction.bounds
            # case 1: make reaction reversible
            if not reaction.reversibility:  # skip if reaction is already reversible
                res = self.check_egc_growth(reaction,m,
                                    (-1000.0,1000.0), # irreversible -> reversible
                                    present_egc, namespace,
                                    compartment)  
                if res:
                    for egc in res:
                        results[egc]["MR"].append(reaction.id)  # MR = make reversible

            # case 2: limit backward reaction
            if original_bounds != (0.0,1000.0):
                res = self.check_egc_growth(reaction,m,
                                    (0.0,1000.0), # Remove Backward (RB) reaction
                                    present_egc, namespace,
                                    compartment)  
                if res:
                    for egc in res:
                        results[egc]["RB"].append(reaction.id)  

            # case 3: limit forward reaction
            if original_bounds != (-1000.0, 0):
                res = self.check_egc_growth(reaction,m,
                                    (-1000.0, 0), # Remove Forward (RF) reaction
                                    present_egc, namespace,
                                    compartment)  
                if res:
                    for egc in res:
                        results[egc]["RF"].append(reaction.id)    

            # case 4: "delete" reaction
            if original_bounds != (0.0, 0.0):
                res = self.check_egc_growth(reaction,m,
                                    (0.0, 0.0),  # Remove (RM) reaction
                                    present_egc, namespace,
                                    compartment)  
                if res:
                    for egc in res:
                        results[egc]["RM"].append(reaction.id)   

        return results



    # @TEST good default for limit and chunksize
    # params not optimal and runtime options need rechecking
    def find_mods_resolve_egcs_greedy(self, model: cobra.Model, 
                                      present_egcs: dict,
                                      namespace:Literal['BiGG']='BiGG',
                                      compartment:list=['c','p']) -> dict:
        """Find the (single) modifications to reactions in a cobra.Model and returns these in a dictionary.
        Splits the modification check in multiple processes.
        
        Args:
            - model (cobra.Model): 
                The model loaded with COBRApy.
            - present_egcs (dict): 
                Dict of the original EGCs found in the model.
            - namespace (Literal['BiGG'], optional): 
                String for the namespace used in the model. 
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional): 
                List of length 2 with the names of the 
                compartments for the dissipations reactions. 
                Defaults to ['c','p'].

        Returns:
            dict: 
                Dictionary of potential modifications to resolve EGCs
                {"egc": {"MR":[potential_solutions],
                "RB":[potential_solutions],
                "RF":[potential_solutions],
                "RM":[potential_solutions]}}
        """

        output_list = []
        
        if len(present_egcs) == 0:
            print('No EGCs present, nothing to solve.') 

        else:
            print("_____________________________________")
            print(f"Try to resolve the following EGCs: {[egc for egc in present_egcs]} \nThis might take a while...")

            # try to find BOF
            pos_bofs = test_biomass_presence(model)
            if pos_bofs:
                bof = pos_bofs[0]
            else:
                mes = 'No growth or biomass objectuve function in model. Cannot solve EGCs.'
                raise KeyError(mes)
            
            with model as m:
                set_bounds_to_default(m)
                m.objective = bof

                # might limit processes with Pool(process=limit) -> otherwise it consumes all cores
                # limit to half of machine cores -> at least for me no speed increment with more cores
                
                pool = Pool(processes=self.limit) 
                            
                # partial -> creates function with fixed variables to call with each iteration
                # needed since pool.imap cannot do that
                part_test_mods = partial(self.test_modifications,
                                        model=m,
                                        present_egc=present_egcs,
                                        namespace=namespace,
                                        compartment=compartment)
                # increment in chunksize will reduce computation time -> but progressbar update also...
                for res in list(tqdm(pool.imap_unordered(func=part_test_mods, iterable=m.reactions, 
                                                        chunksize=self.chunksize),
                                total=len(m.reactions),
                                desc="Resolve EGCs")):
                    if res:
                        output_list.append(res)

                # @TODO : make sure this is run, even if program is discarded
                pool.close()
                pool.join()
                
            # merge output_list to the final results
            results = {}
            for output_dict in output_list:
                for egc, modifications in output_dict.items():
                    for mod, reactions in modifications.items():
                        if not egc in results.keys():
                            results[egc] = {}
                        if mod in results[egc].keys():
                            results[egc][mod] = list(set(results[egc][mod] + reactions))
                        else:
                            results[egc][mod] = list(set(reactions))
                        
            return results


    # find a solution to solve all EGCs that 
    # have a solution based on previous step
    # --------------------------------------
    def find_solution_greedy(self, results:dict, egc_reactions:dict) -> tuple:
        """Based on the originally found EGCs and the output of 
        :py:func:`find_mods_resolve_egcs_greedy`, find a solution that
        solves all EGCs that can be solved with the results.

        Args:
            - results (dict): 
                Output of :py:func:`find_mods_resolve_egcs_greedy`.
            - egc_reactions (dict): 
                Output of :py:meth:`~refinegems.classes.egcs.EGCSolver.find_egcs` with 'with_reac=True'.
                Should be the EGCs before calculating any solutions and applying them.
            - scoring_matrix (dict, optional): 
                Dictionary of the modifications types 
                (RM, MR, RF, RB) and their penality score.
                Defaults to the in-build scoring matrix.

        Returns:
            tuple: 
                Tuple of (1) dict & (2) int:
                
                (1) dictionary of reaction IDs and their mode of change
                (2) score of the solution
        """
    
        solved_egcs = set()
        score = 0
        reacs_for_solu = {}

        # make table and drop EGCs that are not in the model
        solution_table = pd.DataFrame(results)[list(egc_reactions.keys())]
        for col in solution_table.columns:
            solution_table[col] = solution_table[col].apply(lambda x: ', '.join(map(str,x)))
        # find not solvable (with the current code) cyles 
        not_solvable = []
        for col in egc_reactions.keys():
            if all(len(_) == 0 for _ in solution_table[col]):
                not_solvable.append(col)
        # report the problems for manual curation
        if len(not_solvable) > 0:
            print(f'The following EGCs cannot be fixed with the current code: {not_solvable}')
            solution_table.drop(not_solvable, axis=1, inplace=True)
        # fix the EGCs that can be fixed with single changes
        solution_table = solution_table.T
        for egc in solution_table.index.to_list():
            # check if sgc has already been taken care of 
            if egc in solved_egcs:
                continue
        
            # best solution for this cycle
            for mode in self.scoring_matrix.keys():
                if solution_table.loc[egc, mode]:
                    reac = solution_table.loc[egc,mode].split(', ')[0]
                    newly_solved = [_ for _ in solution_table.index[solution_table[mode].str.contains(reac)].to_list() if _ not in solved_egcs]
                    if len(newly_solved) > 0:
                        solved_egcs = set([*solved_egcs,*newly_solved])
                        score += self.scoring_matrix[mode] 
                        reacs_for_solu[reac] = mode

        return (reacs_for_solu, score)


    # apply the found solution the model
    # ----------------------------------
    def apply_modifications(self, model: cobra.Model, solution: dict):
        """Apply the modifications to reactions in solution to the model.

        4 modifications are possible:

        - "RM" -> removes the reaction
        - "RB" -> removes the backwards reaction
        - "RF" -> removes the forward reaction
        - "MR" -> makes reaction reversible

        Args:
            - model (cobra.Model): 
                Input model
            - solution (dict): 
                Best solution from calculation in `py:func:find_solution_greedy`.
        """
        for reac, mode in solution.items():
                reaction = model.reactions.get_by_id(reac)
                if type(reaction) is cobra.Reaction:
                    match mode:  
                        case "RM":
                            reaction.delete()
                        case "RB":
                            reaction.bounds = (0.0, 1000.0)
                        case "RF":
                            model = reaction.bounds = (1000.0, 0.0)
                        case "MR":
                            model = reaction.bounds = (1000.0, 1000.0)
                        case _:
                            mes = f"{mode} is no viable modification... Something went wrong."
                            warnings.warn(mes)

    # run the complete solving process
    # --------------------------------
    def solve_egcs(self, model:cobra.Model, namespace:Literal['BiGG']='BiGG',
                   compartment:list=['c','p']) -> dict|None:
        """Run the complete greedy EGC solving process.

        Note: The input model gets changed, if EGCs can be solved.

        Args:
            - model (cobra.Model): 
                The model loaded with COBRApy.
            - namespace (Literal['BiGG'], optional): 
                String for the namespace used in the model. 
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional): 
                List of length 2 with the names of the 
                compartments for the dissipations reactions. 
                Defaults to ['c','p'].

        Returns:
            dict: 
                Dictionary with the following entries.

                - 'solution': List of reactions for the solution.
                - 'score': Score of the solution.
                - 'remaining egcs': List of EGCs that could not be solved.

        """
        egc_reactions, obj_vals = self.find_egcs(model, with_reacs=True,
                                                 namespace=namespace,
                                                 compartment=compartment)
        results = self.find_mods_resolve_egcs_greedy(model=model, 
                                                     present_egcs=egc_reactions,
                                                     namespace=namespace,
                                                     compartment=compartment)
        if results:
            solution,score = self.find_solution_greedy(results,egc_reactions)
            self.apply_modifications(model,solution)
            still_egc = self.find_egcs(model,namespace=namespace,compartment=compartment)

            return {'solution':solution, 'score':score, 'remaining egcs':still_egc}